import warnings
import difflib
import abc
import os
from difflib import get_close_matches
import atexit
from functools import partial

import sqlalchemy
from sqlalchemy.engine import Engine
from sqlalchemy.exc import (
    NoSuchModuleError,
    OperationalError,
    StatementError,
    PendingRollbackError,
    InternalError,
    ProgrammingError,
)

from sql.run.sparkdataframe import handle_spark_dataframe

from IPython.core.error import UsageError
import sqlglot
import sqlparse


from sql.store import store
from sql import exceptions, display
from sql.error_handler import handle_exception
from sql.parse import (
    escape_string_literals_with_colon_prefix,
    find_named_parameters,
    ConnectionsFile,
)
from sql.warnings import JupySQLQuotedNamedParametersWarning, JupySQLRollbackPerformed
from sql import _current
from sql.connection import error_handling

BASE_DOC_URL = "https://jupysql.ploomber.io/en/latest"


PLOOMBER_DOCS_LINK_STR = f"{BASE_DOC_URL}/connecting.html"

IS_SQLALCHEMY_ONE = int(sqlalchemy.__version__.split(".")[0]) == 1

# Check Full List: https://docs.sqlalchemy.org/en/20/dialects


MISSING_PACKAGE_LIST_EXCEPT_MATCHERS = {
    # SQLite
    "sqlite": "sqlite",
    "pysqlcipher3": "pysqlcipher3",
    # DuckDB
    "duckdb": "duckdb-engine",
    # MySQL + MariaDB
    "pymysql": "pymysql",
    "mysqldb": "mysqlclient",
    "mariadb": "mariadb",
    "mysql": "mysql-connector-python",
    "asyncmy": "asyncmy",
    "aiomysql": "aiomysql",
    "cymysql": "cymysql",
    "pyodbc": "pyodbc",
    # PostgreSQL
    "psycopg2": "psycopg2",
    "psycopg": "psycopg",
    "pg8000": "pg8000",
    "asyncpg": "asyncpg",
    "psycopg2cffi": "psycopg2cffi",
    # Oracle
    "cx_oracle": "cx_oracle",
    "oracledb": "oracledb",
    # MSSQL
    "pyodbc": "pyodbc",
    "pymssql": "pymssql",
    # snowflake
    "snowflake": "snowflake-sqlalchemy",
}

BASE_DRIVERS_URL = f"{BASE_DOC_URL}/howto/db-drivers.html"

DBNAME_2_DOC_LINK = {
    "psycopg2": f"{BASE_DRIVERS_URL}#postgresql",
    "duckdb": f"{BASE_DRIVERS_URL}#duckdb",
}

DIALECT_NAME_SQLALCHEMY_TO_SQLGLOT_MAPPING = {"postgresql": "postgres", "mssql": "tsql"}

# All the DBs and their respective documentation links
DB_DOCS_LINKS = {
    "duckdb": f"{BASE_DOC_URL}/integrations/duckdb.html",
    "mysql": f"{BASE_DOC_URL}/integrations/mysql.html",
    "mssql": f"{BASE_DOC_URL}/integrations/mssql.html",
    "mariadb": f"{BASE_DOC_URL}/integrations/mariadb.html",
    "clickhouse": f"{BASE_DOC_URL}/integrations/clickhouse.html",
    "postgresql": f"{BASE_DOC_URL}/integrations/postgres-connect.html",
    "questdb": f"{BASE_DOC_URL}/integrations/questdb.html",
}


def extract_module_name_from_ModuleNotFoundError(e):
    return e.name


def extract_module_name_from_NoSuchModuleError(e):
    return str(e).split(":")[-1].split(".")[-1]


class ResultSetCollection:
    def __init__(self) -> None:
        self._result_sets = []

    def append(self, result):
        if result in self._result_sets:
            self._result_sets.remove(result)

        self._result_sets.append(result)

    def is_last(self, result):
        # if there are no results, return True to prevent triggering
        # a query in the database
        if not len(self._result_sets):
            return True

        return self._result_sets[-1] is result

    def close_all(self):
        for r in self._result_sets:
            r.close()

        self._result_sets = []

    def __iter__(self):
        return iter(self._result_sets)

    def __len__(self):
        return len(self._result_sets)


def get_missing_package_suggestion_str(e):
    """Provide a better error when a user tries to connect to a database but they're
    missing the database driver
    """
    suggestion_prefix = "To fix it, "

    module_name = None

    if isinstance(e, ModuleNotFoundError):
        module_name = extract_module_name_from_ModuleNotFoundError(e)
    elif isinstance(e, NoSuchModuleError):
        module_name = extract_module_name_from_NoSuchModuleError(e)

    module_name = module_name.lower()

    error_message = (
        suggestion_prefix + "make sure you are using correct driver name:\n"
        "Ref: https://docs.sqlalchemy.org/en/20/core/engines.html#database-urls"
    )

    # Exact match
    suggested_package = MISSING_PACKAGE_LIST_EXCEPT_MATCHERS.get(module_name)

    if suggested_package:
        error_message = (
            suggestion_prefix
            + "run this in your notebook: "
            + error_handling.install_command(suggested_package)
        )
    else:
        # Closely matched
        close_matches = difflib.get_close_matches(
            module_name, MISSING_PACKAGE_LIST_EXCEPT_MATCHERS.keys()
        )

        if close_matches:
            error_message = (
                f'Perhaps you meant to use driver the dialect: "{close_matches[0]}"'
            )

    error_suffix = (
        DBNAME_2_DOC_LINK.get(module_name)
        if DBNAME_2_DOC_LINK.get(module_name)
        else PLOOMBER_DOCS_LINK_STR
    )

    return error_message + "\n\nFor more details, see: " + error_suffix


def rough_dict_get(dct, sought, default=None):
    """
    Like dct.get(sought), but any key containing sought will do.

    If there is a `@` in sought, seek each piece separately.
    This lets `me@server` match `me:***@myserver/db`
    """

    sought = sought.split("@")
    for key, val in dct.items():
        if not any(s.lower() not in key.lower() for s in sought):
            return val
    return default


def _error_invalid_connection_info(e, connect_str):
    err = UsageError(
        "An error happened while creating the connection: "
        f"{e}.{_suggest_fix(env_var=False, connect_str=connect_str)}"
    )
    err.modify_exception = True
    return err


class ConnectionManager:
    """A class to manage and create database connections"""

    # all connections
    connections = {}

    # the active connection
    current = None

    @classmethod
    def set(
        cls,
        descriptor,
        displaycon,
        connect_args=None,
        creator=None,
        alias=None,
        config=None,
    ):
        """
        Set the current database connection. This method is called from the magic to
        determine which connection to use (either use an existing one or open a new one)

        Parameters
        ----------
        descriptor : str or sqlalchemy.engine.Engine or sqlalchemy.engine.Connection
            A connection string or an existing connection. It opens a new connection
            if needed, otherwise it just assigns the connection as the current
            connection.

        alias : str, optional
            A name to identify the connection

        config : object, optional
            An object with configuration options. Options must be accessible via
            attributes. As of 0.9.0, only the autocommit option is needed.
        """
        connect_args = connect_args or {}

        if descriptor:
            if isinstance(descriptor, SQLAlchemyConnection):
                cls.current = descriptor
            elif isinstance(descriptor, Engine):
                cls.current = SQLAlchemyConnection(
                    descriptor, config=config, alias=alias
                )
            elif is_pep249_compliant(descriptor):
                cls.current = DBAPIConnection(descriptor, config=config, alias=alias)
            elif is_spark(descriptor):
                cls.current = SparkConnectConnection(
                    descriptor, config=config, alias=alias
                )
            else:
                existing = rough_dict_get(cls.connections, descriptor)
                if existing and existing.alias == alias:
                    cls.current = existing
                elif existing and alias is None:
                    if (
                        _current._config_feedback_normal_or_more()
                        and cls.current != existing
                    ):
                        display.message(f"Switching to connection {descriptor!r}")
                    cls.current = existing

                # passing the same URL but different alias: create a new connection
                elif existing is None or existing.alias != alias:
                    is_connect_and_switch, is_connect = False, False
                    if cls.current and cls.current.alias != alias:
                        is_connect_and_switch = True
                    else:
                        is_connect = True

                    cls.current = cls.from_connect_str(
                        connect_str=descriptor,
                        connect_args=connect_args,
                        creator=creator,
                        alias=alias,
                        config=config,
                    )
                    if _current._config_feedback_normal_or_more():
                        identifier = alias or cls.current.url
                        if is_connect_and_switch:
                            display.message(
                                f"Connecting and switching to connection {identifier!r}"
                            )
                        if is_connect:
                            display.message(f"Connecting to {identifier!r}")

        else:
            if cls.connections:
                if displaycon and _current._config_feedback_normal_or_more():
                    cls.display_current_connection()
            elif os.getenv("DATABASE_URL"):
                cls.current = cls.from_connect_str(
                    connect_str=os.getenv("DATABASE_URL"),
                    connect_args=connect_args,
                    creator=creator,
                    alias=alias,
                    config=config,
                )
            else:
                raise cls._error_no_connection()

        return cls.current

    @classmethod
    def close_all(cls, verbose=False):
        """Close all connections"""
        connections = ConnectionManager.connections.copy()
        for name, conn in connections.items():
            conn.close()

            if verbose:
                display.message(f"Closing {name}")

        cls.connections = {}

    @classmethod
    def _error_no_connection(cls):
        """Error when there isn't any connection"""
        err = UsageError("No active connection." + _suggest_fix(env_var=True))
        err.modify_exception = True
        return err

    @classmethod
    def display_current_connection(cls):
        for conn in cls._get_connections():
            if conn["current"]:
                alias = conn.get("alias")
                if alias:
                    display.message(f"Running query in {alias!r}")
                else:
                    display.message(f"Running query in {conn['url']!r}")

    @classmethod
    def _get_connections(cls):
        """
        Return a list of dictionaries
        """
        connections = []

        for key in sorted(cls.connections):
            conn = cls.connections[key]

            is_current = conn == cls.current

            connections.append(
                {
                    "current": is_current,
                    "key": key,
                    "url": conn.url,
                    "alias": conn.alias,
                    "connection": conn,
                }
            )

        return connections

    @classmethod
    def close_connection_with_descriptor(cls, descriptor):
        """Close a connection with the given descriptor"""
        if isinstance(descriptor, SQLAlchemyConnection):
            conn = descriptor
        else:
            conn = cls.connections.get(descriptor) or cls.connections.get(
                descriptor.lower()
            )

        if not conn:
            raise exceptions.RuntimeError(
                "Could not close connection because it was not found amongst these: %s"
                % str(list(cls.connections.keys()))
            )

        if descriptor in cls.connections:
            cls.connections.pop(descriptor)
        else:
            cls.connections.pop(
                str(conn.metadata.bind.url) if IS_SQLALCHEMY_ONE else str(conn.url)
            )

        conn.close()

    @classmethod
    def connections_table(cls):
        """Returns the current connections as a table"""
        connections = cls._get_connections()

        def map_values(d):
            d["current"] = "*" if d["current"] else ""
            d["alias"] = d["alias"] if d["alias"] else ""
            return d

        return display.ConnectionsTable(
            headers=["current", "url", "alias"],
            rows_maps=[map_values(c) for c in connections],
        )

    @classmethod
    def from_connect_str(
        cls, connect_str=None, connect_args=None, creator=None, alias=None, config=None
    ):
        """Creates a new connection from a connection string"""
        connect_args = connect_args or {}

        try:
            if creator:
                engine = sqlalchemy.create_engine(
                    connect_str,
                    connect_args=connect_args,
                    creator=creator,
                )
            else:
                engine = sqlalchemy.create_engine(
                    connect_str,
                    connect_args=connect_args,
                )
        except (ModuleNotFoundError, NoSuchModuleError) as e:
            suggestion_str = get_missing_package_suggestion_str(e)
            raise exceptions.MissingPackageError(
                "\n\n".join([str(e), suggestion_str])
            ) from e
        except Exception as e:
            raise _error_invalid_connection_info(e, connect_str) from e

        connection = SQLAlchemyConnection(engine, alias=alias, config=config)
        connection.connect_args = connect_args

        return connection

    @classmethod
    def load_default_connection_from_file_if_any(cls, config):
        try:
            connections_file = ConnectionsFile(path_to_file=config.dsn_filename)
        except FileNotFoundError:
            return

        default_url = connections_file.get_default_connection_url()

        if default_url is not None:
            try:
                cls.set(
                    default_url,
                    displaycon=False,
                    alias="default",
                    config=config,
                )
            except Exception as e:
                # this is executed during the magic initialization, we don't want
                # to raise an exception here because it would prevent the magic
                # from being used
                display.message_warning(
                    "WARNING: Cannot start default connection from .ini file:"
                    f"\n\n{str(e)}"
                )


class AbstractConnection(abc.ABC):
    """The abstract base class for all connections"""

    def __init__(self, alias):
        self.alias = alias

        ConnectionManager.current = self
        ConnectionManager.connections[alias] = self

        self._result_sets = ResultSetCollection()

    @abc.abstractproperty
    def dialect(self):
        """Returns a string with the SQL dialect name"""
        pass

    @abc.abstractmethod
    def raw_execute(self, query, parameters=None):
        """Run the query without any pre-processing"""
        pass

    @abc.abstractmethod
    def _get_database_information(self):
        """
        Get the dialect, driver, and database server version info of current
        connection
        """
        pass

    @abc.abstractmethod
    def to_table(self, table_name, data_frame, if_exists, index, schema=None):
        """Create a table from a pandas DataFrame"""
        pass

    def close(self):
        """Close the connection"""
        for rs in self._result_sets:
            # this might be None if it didn't run any query
            if rs._sqlaproxy is not None:
                rs._sqlaproxy.close()

        self._connection.close()

    def _get_sqlglot_dialect(self):
        """
        Get the sqlglot dialect, this is similar to the dialect property except it
        maps some dialects to their sqlglot equivalent. This method should only be
        used for the transpilation process, for any other purposes, use the dialect
        property.

        Returns
        -------
        str
            Available dialect in sqlglot package, see more:
            https://github.com/tobymao/sqlglot/blob/main/sqlglot/dialects/dialect.py
        """
        connection_info = self._get_database_information()
        return DIALECT_NAME_SQLALCHEMY_TO_SQLGLOT_MAPPING.get(
            connection_info["dialect"], connection_info["dialect"]
        )

    def _transpile_query(self, query):
        """Translate the given SQL clause that's compatible to current connected
        dialect by sqlglot

        Parameters
        ----------
        query : str
            Original SQL clause

        Returns
        -------
        str
            SQL clause that's compatible to current connected dialect
        """
        write_dialect = self._get_sqlglot_dialect()

        # we write queries to be duckdb-compatible so we don't need to transpile
        # them. Furthermore, sqlglot does not guarantee roundtrip conversion
        # so calling transpile might break queries
        if write_dialect == "duckdb":
            return query

        try:
            return ";\n".join(
                [p.sql(dialect=write_dialect) for p in sqlglot.parse(query)]
            )
        except Exception:
            return query

    def _prepare_query(self, query, with_=None) -> str:
        """
        Returns a textual representation of a query based
        on the current connection

        Parameters
        ----------
        query : str
            SQL query

        with_ : string, default None
            The key to use in with sql clause
        """
        if with_:
            query = self._resolve_cte(query, with_)

        query = self._transpile_query(query)

        return query

    def _resolve_cte(self, query, with_):
        return str(store.render(query, with_=with_))

    def execute(self, query, with_=None):
        """
        Executes SQL query on a given connection
        """
        query_prepared = self._prepare_query(query, with_)
        return self.raw_execute(query_prepared)

    def is_use_backtick_template(self):
        """Get if the dialect support backtick (`) syntax as identifier

        Returns
        -------
        bool
            Indicate if the dialect can use backtick identifier in the SQL clause
        """
        cur_dialect = self._get_sqlglot_dialect()
        if not cur_dialect:
            return False
        try:
            return (
                "`" in sqlglot.Dialect.get_or_raise(cur_dialect).Tokenizer.IDENTIFIERS
            )
        except (ValueError, AttributeError, TypeError):
            return False

    def get_curr_identifiers(self) -> list:
        """
        Returns list of identifiers for current connection

        Default identifiers are : ["", '"']
        """
        identifiers = ["", '"']
        try:
            connection_info = self._get_database_information()
            if connection_info:
                cur_dialect = connection_info["dialect"]
                identifiers_ = sqlglot.Dialect.get_or_raise(
                    cur_dialect
                ).Tokenizer.IDENTIFIERS

                identifiers = [*set(identifiers + identifiers_)]
        except ValueError:
            pass
        except AttributeError:
            # this might be a DBAPI connection
            pass

        return identifiers


# some dialects break when commit is used
_COMMIT_BLACKLIST_DIALECTS = (
    "athena",
    "bigquery",
    "clickhouse",
    "ingres",
    "mssql",
    "teradata",
    "vertica",
)


# TODO: the autocommit is read only during initialization, if the user changes it
# it won't have any effect
class SQLAlchemyConnection(AbstractConnection):
    """Manages connections to databases

    Parameters
    ----------
    engine: sqlalchemy.engine.Engine
        The SQLAlchemy engine to use
    """

    is_dbapi_connection = False

    def __init__(self, engine, alias=None, config=None):
        if IS_SQLALCHEMY_ONE:
            self._metadata = sqlalchemy.MetaData(bind=engine)
        else:
            self._metadata = None

        # this returns a url with the password replaced by ***
        self._url = (
            repr(sqlalchemy.MetaData(bind=engine).bind.url)
            if IS_SQLALCHEMY_ONE
            else repr(engine.url)
        )

        self._connection_sqlalchemy = self._start_sqlalchemy_connection(
            engine, self._url
        )

        db_info = self._get_database_information()
        self._dialect = db_info["dialect"]
        self._driver = db_info["driver"]

        autocommit = True if config is None else config.autocommit

        if autocommit:
            success = set_sqlalchemy_isolation_level(self._connection_sqlalchemy)
            self._requires_manual_commit = not success

            # TODO: I noticed we don't have any unit tests for this
            # even if autocommit is true, we should not use it for some dialects
            self._requires_manual_commit = (
                all(
                    blacklisted_dialect not in str(self._dialect)
                    for blacklisted_dialect in _COMMIT_BLACKLIST_DIALECTS
                )
                and self._requires_manual_commit
            )
        else:
            self._requires_manual_commit = False

        # TODO: we're no longer using this. I believe this is only used via the
        # config.feedback option
        self.name = default_alias_for_engine(engine)

        # calling init from AbstractConnection must be the last thing we do as it
        # register the connection
        super().__init__(alias=alias or self._url)

    @property
    def dialect(self):
        return self._dialect

    @property
    def driver(self):
        return self._driver

    def _connection_execute(self, query, parameters=None):
        """Call the connection execute method

        Parameters
        ----------
        query : str
            SQL query

        parameters : dict, default None
            Parameters to use in the query (:variable format)
        """
        # we do not support multiple statements
        if len(sqlparse.split(query)) > 1:
            raise NotImplementedError("Only one statement is supported.")

        operation = partial(self._execute_with_parameters, query, parameters)
        out = self._execute_with_error_handling(operation)

        if self._requires_manual_commit:
            # Calling connection.commit() when using duckdb-engine will yield
            # empty results if we commit after a SELECT or SUMMARIZE statement,
            # see: https://github.com/Mause/duckdb_engine/issues/734.
            if self.dialect == "duckdb":
                no_commit = detect_duckdb_summarize_or_select(query)
                if no_commit:
                    return out

            # in sqlalchemy 1.x, connection has no commit attribute
            if IS_SQLALCHEMY_ONE:
                # TODO: I moved this from run.py where we were catching all exceptions
                # because some drivers do not support commits. However, I noticed
                # that if I remove the try catch we get this error in SQLite:
                #  "cannot commit - no transaction is active", we need to investigate
                # further, catching generic exceptions is not a good idea
                try:
                    self._connection.execute("commit")
                except Exception:
                    pass
            else:
                self._connection.commit()

        return out

    def _execute_with_parameters(self, query, parameters):
        """Execute the query with the given parameters"""
        if parameters == {}:
            return self._connection.exec_driver_sql(query)

        parameters = parameters or {}
        if IS_SQLALCHEMY_ONE:
            out = self._connection.execute(sqlalchemy.text(query), **parameters)
        else:
            out = self._connection.execute(
                sqlalchemy.text(query), parameters=parameters
            )

        return out

    def raw_execute(self, query, parameters=None, with_=None):
        """Run the query without any preprocessing

        Parameters
        ----------
        query : str
            SQL query

        parameters : dict, default None
            Parameters to use in the query. They should appear in the query with the
            :name format (no quotes around them)

        with_ : list, default None
            List of CTEs to use in the query
        """
        # mssql with pyodbc does not support multiple open result sets, so we need
        # to close them all before issuing a new query
        if self.dialect == "mssql" and self.driver == "pyodbc":
            self._result_sets.close_all()

        if with_:
            query = self._resolve_cte(query, with_)

        query, quoted_named_parameters = escape_string_literals_with_colon_prefix(query)

        if quoted_named_parameters and parameters:
            intersection = set(quoted_named_parameters) & set(parameters)

            if intersection:
                intersection_ = ", ".join(sorted(intersection))
                warnings.warn(
                    f"The following variables are defined: {intersection_}. However "
                    "the parameters are quoted in the query, if you want to use "
                    "them as named parameters, remove the quotes.",
                    category=JupySQLQuotedNamedParametersWarning,
                )

        if parameters:
            required_parameters = set(sqlalchemy.text(query).compile().params)
            available_parameters = set(parameters)
            missing_parameters = required_parameters - available_parameters

            if missing_parameters:
                raise exceptions.InvalidQueryParameters(
                    "Cannot execute query because the following "
                    "variables are undefined: {}".format(", ".join(missing_parameters))
                )

            return self._connection_execute(query, parameters)
        else:
            try:
                return self._connection_execute(query, parameters)
            except StatementError as e:
                # add a more helpful message if the users passes :variable but
                # the feature isn't enabled
                if parameters is None:
                    named_params = find_named_parameters(query)

                    if named_params:
                        named_params_ = ", ".join(named_params)
                        e.add_detail(
                            f"Your query contains named parameters ({named_params_}) "
                            'but the named parameters feature is "warn". \nEnable it '
                            'with: %config SqlMagic.named_parameters="enabled" \nor '
                            "disable it with: "
                            '%config SqlMagic.named_parameters="disabled"\n'
                            "For more info, see the docs: "
                            "https://jupysql.ploomber.io/en/latest/api/configuration.html#named-parameters"  # noqa
                        )
                elif parameters == {}:
                    e.add_detail(
                        'The named parameters feature is "disabled". '
                        'Enable it with: %config SqlMagic.named_parameters="enabled".\n'
                        "For more info, see the docs: "
                        "https://jupysql.ploomber.io/en/latest/api/configuration.html#named-parameters"  # noqa
                    )
                raise

    def _execute_with_error_handling(self, operation):
        """Execute a database operation and handle errors

        Parameters
        ----------
        operation : callable
            A callable that takes no parameters to execute a database operation
        """
        rollback_needed = False

        try:
            out = operation()

        # this is a generic error but we've seen it in postgres. it helps recover
        # from a idle session timeout (happens in psycopg 2 and psycopg 3)
        except PendingRollbackError:
            warnings.warn(
                "Found invalid transaction. JupySQL executed a ROLLBACK operation.",
                category=JupySQLRollbackPerformed,
            )
            rollback_needed = True

        # postgres error
        except InternalError as e:
            # message from psycopg 2 and psycopg 3
            message = (
                "current transaction is aborted, "
                "commands ignored until end of transaction block"
            )
            if type(e.orig).__name__ == "InFailedSqlTransaction" and message in str(
                e.orig
            ):
                warnings.warn(
                    (
                        "Current transaction is aborted. "
                        "JupySQL executed a ROLLBACK operation."
                    ),
                    category=JupySQLRollbackPerformed,
                )
                rollback_needed = True
            else:
                raise

        # postgres error
        except OperationalError as e:
            # message from psycopg 2 and psycopg 3
            message = "server closed the connection unexpectedly"

            if type(e.orig).__name__ == "OperationalError" and message in str(e.orig):
                warnings.warn(
                    "Server closed connection. JupySQL executed a ROLLBACK operation.",
                    category=JupySQLRollbackPerformed,
                )
                rollback_needed = True
            else:
                raise

        except ProgrammingError as e:
            # error when accessing previously non-existing file with duckdb using
            # sqlalchemy 2.x
            if "duckdb.InvalidInputException" in str(e) and "please ROLLBACK" in str(e):
                rollback_needed = True
            else:
                raise

        if rollback_needed:
            self._connection.rollback()
            out = operation()

        return out

    def _get_database_information(self):
        dialect = self._connection_sqlalchemy.dialect

        return {
            "dialect": getattr(dialect, "name", None),
            "driver": getattr(dialect, "driver", None),
            # NOTE: this becomes available after calling engine.connect()
            "server_version_info": getattr(dialect, "server_version_info", None),
        }

    @property
    def url(self):
        """Returns an obfuscated connection string (password hidden)"""
        return self._url

    @property
    def connection_sqlalchemy(self):
        """Returns the SQLAlchemy connection object"""
        return self._connection_sqlalchemy

    @property
    def _connection(self):
        """Returns the SQLAlchemy connection object"""
        return self._connection_sqlalchemy

    def close(self):
        super().close()

        # NOTE: in SQLAlchemy 2.x, we need to call engine.dispose() to completely
        # close the connection, calling connection.close() is not enough
        self._connection.engine.dispose()

    @classmethod
    def _start_sqlalchemy_connection(cls, engine, connect_str):
        try:
            connection = engine.connect()
            return connection
        except OperationalError as e:
            handle_exception(e)
        except Exception as e:
            raise _error_invalid_connection_info(e, connect_str) from e

    def to_table(self, table_name, data_frame, if_exists, index, schema=None):
        """Create a table from a pandas DataFrame"""
        operation = partial(
            data_frame.to_sql,
            table_name,
            self.connection_sqlalchemy,
            if_exists=if_exists,
            index=index,
            schema=schema,
        )

        try:
            self._execute_with_error_handling(operation)
        except ValueError:
            raise exceptions.ValueError(
                f"Table {table_name!r} already exists. Consider using "
                "--persist-replace to drop the table before "
                "persisting the data frame"
            )

        display.message_success(f"Success! Persisted {table_name} to the database.")


class DBAPIConnection(AbstractConnection):
    """A connection object for generic DBAPI connections"""

    is_dbapi_connection = True

    def __init__(self, connection, alias=None, config=None):
        # detect if the engine is a native duckdb connection
        self._is_duckdb_native = _check_if_duckdb_dbapi_connection(connection)

        self._dialect = "duckdb" if self._is_duckdb_native else None
        self._driver = None

        # TODO: implement the dialect blacklist and add unit tests
        self._requires_manual_commit = True if config is None else config.autocommit

        self._connection = connection
        self._connection_class_name = type(connection).__name__

        # calling init from AbstractConnection must be the last thing we do as it
        # register the connection
        super().__init__(alias=alias or self._connection_class_name)

        # TODO: delete this
        self.name = self._connection_class_name

    @property
    def dialect(self):
        return self._dialect

    @property
    def driver(self):
        return self._driver

    def raw_execute(self, query, parameters=None, with_=None):
        """Run the query without any preprocessing

        Parameters
        ----------
        query : str
            SQL query

        parameters : dict, default None
            This parameter is added for consistency with SQLAlchemy connections but
            it is not used
        """
        # we do not support multiple statements (this might actually work in some
        # drivers but we need to add this for consistency with SQLAlchemyConnection)
        if len(sqlparse.split(query)) > 1:
            raise NotImplementedError("Only one statement is supported.")

        if with_:
            query = self._resolve_cte(query, with_)

        cur = self._connection.cursor()

        # NOTE: this is a workaround for duckdb 1.1.0 and higher so we keep the
        # existing behavior of being able to query data frames
        if self._is_duckdb_native:
            try:
                cur.execute("SET python_scan_all_frames=true")
            except Exception:
                pass

        cur.execute(query)

        if self._requires_manual_commit:
            self._connection.commit()

        return cur

    def _get_database_information(self):
        return {
            "dialect": self.dialect,
            "driver": self._connection_class_name,
            "server_version_info": None,
        }

    @property
    def url(self):
        """Returns None since DBAPI connections don't have a url"""
        return None

    @property
    def connection_sqlalchemy(self):
        """
        Raises NotImplementedError since DBAPI connections don't have a SQLAlchemy
        connection object
        """
        raise NotImplementedError(
            "This feature is only available for SQLAlchemy connections"
        )

    def to_table(self, table_name, data_frame, if_exists, index, schema=None):
        raise exceptions.NotImplementedError(
            "--persist/--persist-replace is not available for DBAPI connections"
            " (only available for SQLAlchemy connections)"
        )


class SparkConnectConnection(AbstractConnection):
    is_dbapi_connection = False

    def __init__(self, connection, alias=None, config=None):
        self._driver = None

        # TODO: implement the dialect blacklist and add unit tests
        self._requires_manual_commit = True if config is None else config.autocommit

        self._connection = connection
        self._connection_class_name = type(connection).__name__

        # calling init from AbstractConnection must be the last thing we do as it
        # register the connection
        super().__init__(alias=alias or self._connection_class_name)

        self.name = self._connection_class_name

    @property
    def dialect(self):
        """Returns a string with the SQL dialect name"""
        return "spark2"

    def raw_execute(self, query, parameters=None):
        """Run the query without any pre-processing"""
        return handle_spark_dataframe(self._connection.sql(query))

    def _get_database_information(self):
        """
        Get the dialect, driver, and database server version info of current
        connection
        """
        return {
            "dialect": self.dialect,
            "driver": self._connection_class_name,
            "server_version_info": self._connection.version,
        }

    @property
    def url(self):
        """Returns None since Spark connections don't have a url"""
        return None

    @property
    def connection_sqlalchemy(self):
        """
        Raises NotImplementedError since Spark connections don't have a SQLAlchemy
        connection object
        """
        raise NotImplementedError(
            "This feature is only available for SQLAlchemy connections"
        )

    def to_table(self, table_name, data_frame, if_exists, index, schema=None):
        mode = (
            "overwrite"
            if if_exists == "replace"
            else "append" if if_exists == "append" else "error"
        )
        self._connection.createDataFrame(data_frame).write.mode(mode).saveAsTable(
            f"{schema}.{table_name}" if schema else table_name
        )

    def close(self):
        """Override of the abstract close as SparkSession is usually
        shared with pyspark"""
        pass


def _check_if_duckdb_dbapi_connection(conn):
    """Check if the connection is a native duckdb connection"""
    # NOTE: duckdb defines df and pl to efficiently convert results to
    # pandas.DataFrame and polars.DataFrame respectively
    return hasattr(conn, "df") and hasattr(conn, "pl")


def _suggest_fix(env_var, connect_str=None):
    """
    Returns an error message that we can display to the user
    to tell them how to pass the connection string
    """
    DEFAULT_PREFIX = "\n\n"
    prefix = ""

    if connect_str:
        matches = get_close_matches(
            connect_str, list(ConnectionManager.connections), n=1
        )
        matches_db = get_close_matches(
            connect_str.lower(), list(DB_DOCS_LINKS.keys()), cutoff=0.3, n=1
        )

        if matches:
            prefix = prefix + (
                "\n\nPerhaps you meant to use the existing "
                f"connection: %sql {matches[0]!r}?"
            )

        if matches_db:
            prefix = prefix + (
                f"\n\nPerhaps you meant to use the {matches_db[0]!r} db \n"
                f"To find more information regarding connection: "
                f"{DB_DOCS_LINKS[matches_db[0]]}\n\n"
            )

        if not matches and not matches_db:
            prefix = DEFAULT_PREFIX
    else:
        matches = None
        matches_db = None
        prefix = DEFAULT_PREFIX

    connection_string = (
        "Pass a valid connection string:\n    "
        "Example: %sql postgresql://username:password@hostname/dbname"
    )

    suffix = "To fix it:" if not matches else "Otherwise, try the following:"
    options = [f"{prefix}{suffix}", connection_string]

    keys = list(ConnectionManager.connections.keys())

    if keys:
        keys_ = ",".join(repr(k) for k in keys)
        options.append(
            f"Pass a connection key (one of: {keys_})"
            f"\n    Example: %sql {keys[0]!r}"
        )

    if env_var:
        options.append("Set the environment variable $DATABASE_URL")

    if len(options) >= 3:
        options.insert(-1, "OR")

    options.append(f"For more details, see: {PLOOMBER_DOCS_LINK_STR}")

    return "\n\n".join(options)


def is_pep249_compliant(conn):
    """
    Checks if given connection object complies with PEP 249
    """
    pep249_methods = [
        "close",
        "commit",
        # "rollback",
        # "cursor",
        # PEP 249 doesn't require the connection object to have
        # a cursor method strictly
        # ref: https://peps.python.org/pep-0249/#id52
    ]

    for method_name in pep249_methods:
        # Checking whether the connection object has the method
        # and if it is callable
        if not hasattr(conn, method_name) or not callable(getattr(conn, method_name)):
            return False

    return True


def is_spark(conn):
    """Check if it is a SparkSession by checking for available methods"""

    sparksession_methods = [
        "table",
        "read",
        "createDataFrame",
        "sql",
        "stop",
        "catalog",
        "version",
    ]
    for method_name in sparksession_methods:
        # Checking whether the connection object has the method
        if not hasattr(conn, method_name):
            return False

    return True


def default_alias_for_engine(engine):
    if not engine.url.username:
        # keeping this for compatibility
        return str(engine.url)

    return f"{engine.url.username}@{engine.url.database}"


def set_sqlalchemy_isolation_level(conn):
    """
    Sets the autocommit setting for a database connection using SQLAlchemy.
    This better handles some edge cases than calling .commit() on the connection but
    not all drivers support it.
    """
    try:
        conn.execution_options(isolation_level="AUTOCOMMIT")
        return True
    except Exception:
        return False


def detect_duckdb_summarize_or_select(query):
    """
    Checks if the SQL query is a DuckDB SELECT or SUMMARIZE statement.

    Note:
    Assumes there is only one SQL statement in the query.
    """
    statements = sqlparse.parse(query)
    if statements:
        if len(statements) > 1:
            raise NotImplementedError("Multiple statements are not supported")
        stype = statements[0].get_type()
        if stype == "SELECT":
            return True
        elif stype == "UNKNOWN":
            # Further analysis is required
            sql_stripped = sqlparse.format(query, strip_comments=True)
            words = sql_stripped.split()
            return len(words) > 0 and (
                words[0].lower() == "from" or words[0].lower() == "summarize"
            )
    return False


atexit.register(ConnectionManager.close_all, verbose=True)



import os.path
import codecs
import csv
from io import StringIO


class CSVWriter:
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        _row = row
        self.writer.writerow(_row)
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)
        self.queue.seek(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)


class CSVResultDescriptor:
    """
    Provides IPython Notebook-friendly output for the
    feedback after a ``.csv`` called.
    """

    def __init__(self, file_path):
        self.file_path = file_path

    def __repr__(self):
        return "CSV results at %s" % os.path.join(os.path.abspath("."), self.file_path)

    def _repr_html_(self):
        return '<a href="%s">CSV results</a>' % os.path.join(
            ".", "files", self.file_path
        )



import sys
import types


def show_usage_error(self, exc):
    """
    This is a patched version of IPython's InteractiveShell.show_usage_error,
    which allows us to pass a custom prefix in the error message.
    """
    if hasattr(exc, "error_type"):
        print(f"{exc.error_type}: {exc}", file=sys.stderr)
    else:
        print(f"UsageError: {exc}", file=sys.stderr)


def patch_ipython_usage_error(ip):
    """Patch IPython so we can customize UsageError's messages"""
    ip.show_usage_error = types.MethodType(show_usage_error, ip)



import re
import itertools
import shlex
from os.path import expandvars
from pathlib import Path
import configparser
import warnings
import ast

from sqlalchemy.engine.url import URL

from sql import exceptions
from sql.util import check_duplicate_arguments

# Keywords used to identify the beginning of SQL queries
# in split_args_and_sql(). Should cover all cases but can
# be edited to include special keywords.
SQL_COMMANDS = [
    "select",
    "from",
    "with",
    "pivot",
    "create",
    "update",
    "delete",
    "insert",
    "alter",
    "drop",
    "describe",
]


def _parse_config_section(section):
    """Return a given configuration section as a dictionary of keys and values

    If the section contains `query` as key, its value is evaluated such
    that a `"{...}"` string is also converted to a dictionary.

    Parameters
    ----------
    section : list[tuple[str,str]]
        The section object as returned by ConfigParser.items()
    """
    url_args = dict(section)

    if "query" in url_args:
        url_args["query"] = ast.literal_eval(url_args["query"])

    return url_args


class ConnectionsFile:
    def __init__(self, path_to_file) -> None:
        self.parser = configparser.ConfigParser()
        dsn_file = Path(path_to_file)

        cfg_content = dsn_file.read_text()
        self.parser.read_string(cfg_content)

    def get_default_connection_url(self):
        try:
            section = self.parser.items("default")
        except configparser.NoSectionError:
            return None

        url = URL.create(**_parse_config_section(section))
        return str(url.render_as_string(hide_password=False))


def connection_str_from_dsn_section(section, config):
    """Return a SQLAlchemy connection string from a section in a DSN file

    Parameters
    ----------
    section : str
        The section name in the DSN file

    config : Config
        The config object, must have a dsn_filename attribute
    """
    parser = configparser.ConfigParser()
    dsn_file = Path(config.dsn_filename)

    try:
        cfg_content = dsn_file.read_text()
    except FileNotFoundError as e:
        raise exceptions.FileNotFoundError(
            f"%config SqlMagic.dsn_filename ({str(config.dsn_filename)!r}) not found."
            " Ensure the file exists or change the configuration: "
            "%config SqlMagic.dsn_filename = 'path/to/file.ini'"
        ) from e

    try:
        parser.read_string(cfg_content)
    except configparser.Error as e:
        raise exceptions.RuntimeError(
            "An error happened when loading "
            "your %config SqlMagic.dsn_filename "
            f"({config.dsn_filename!r})\n{type(e).__name__}: {e}"
        ) from e

    try:
        cfg = parser.items(section)
    except configparser.NoSectionError as e:
        raise exceptions.KeyError(
            f"The section {section!r} does not exist in the "
            f"connections file {config.dsn_filename!r}"
        ) from e

    try:
        url = URL.create(**_parse_config_section(cfg))
    except TypeError as e:
        if "unexpected keyword argument" in str(e):
            raise exceptions.TypeError(
                f"%config SqlMagic.dsn_filename ({config.dsn_filename!r}) is invalid. "
                "It must only contain the following keys: drivername, username, "
                "password, host, port, database, query"
            ) from e
        else:
            raise

    return str(url.render_as_string(hide_password=False))


def _connection_string(arg, path_to_file):
    """
    Given a string, return a SQLAlchemy connection string if possible.

    Scenarios:

    - If the string is a valid URL, return it
    - If the string is a valid section in the DSN file return the connection string
    - Otherwise return an empty string

    Parameters
    ----------
    arg : str
        The string to parse

    path_to_file : str
        The path to the DSN file
    """
    # for environment variables
    arg = expandvars(arg)

    # if it's a URL, return it
    if "@" in arg or "://" in arg:
        return arg

    # if it's a section in the DSN file, return the connection string
    if arg.startswith("[") and arg.endswith("]"):
        section = arg.lstrip("[").rstrip("]")
        parser = configparser.ConfigParser()
        parser.read(path_to_file)
        cfg = parser.items(section)
        url = URL.create(**_parse_config_section(cfg))
        url_ = str(url.render_as_string(hide_password=False))

        warnings.warn(
            "Starting connections with: %sql [section_name] is deprecated "
            "and will be removed in a future release. "
            "Please use: %sql --section section_name instead.",
            category=FutureWarning,
        )

        return url_

    return ""


def parse(arg, path_to_file):
    """Extract connection info and result variable from SQL

    Please don't add any more syntax requiring
    special parsing.
    Instead, add @arguments to SqlMagic.execute.

    We're grandfathering the
    connection string and `<<` operator in.

    Parameters
    ----------
    arg : str
        The string to parse

    path_to_file : str
        The path to the DSN file
    """
    result = {
        "connection": "",
        "sql": "",
        "result_var": None,
        "return_result_var": False,
    }

    pieces = arg.split(None, 1)
    if not pieces:
        return result

    result["connection"] = _connection_string(pieces[0], path_to_file)

    if result["connection"]:
        if len(pieces) == 1:
            return result
        arg = pieces[1]

    pointer = arg.find("<<")
    if pointer != -1:
        left = arg[:pointer].replace(" ", "").replace("\n", "")
        right = arg[pointer + 2 :].strip(" ")

        if "=" in left:
            result["result_var"] = left[:-1]
            result["return_result_var"] = True
        else:
            result["result_var"] = left

        result["sql"] = right
    else:
        result["sql"] = arg
    return result


def _option_strings_from_parser(parser):
    """Extracts the expected option strings (-a, --append, etc) from argparse parser

    Thanks Martijn Pieters
    https://stackoverflow.com/questions/28881456/how-can-i-list-all-registered-arguments-from-an-argumentparser-instance

    :param parser: [description]
    :type parser: IPython.core.magic_arguments.MagicArgumentParser
    """
    opts = [a.option_strings for a in parser._actions]
    return list(itertools.chain.from_iterable(opts))


def without_sql_comment(parser, line):
    """Strips -- comment from a line

    The argparser unfortunately expects -- to precede an option,
    but in SQL that delineates a comment.  So this removes comments
    so a line can safely be fed to the argparser.

    :param line: A line of SQL, possibly mixed with option strings
    :type line: str
    """

    args = _option_strings_from_parser(parser)
    result = itertools.takewhile(
        lambda word: (not word.startswith("--")) or (word in args),
        shlex.split(line, posix=False),
    )
    return " ".join(result)


def split_args_and_sql(line):
    """Separates line into args and sql query

    The argparser expects - to precede an argument, but postgreSQL
    and duckDB allow for -> and ->> to be used as JSON operators.
    This function splits the line into two - args and sql.
    This way we can only pass the args into the argparser, and
    add in the sql later.

    Parameters
    ----------
    line: str
        A line of SQL, preceded by option/argument strings

    Returns
    -------
    arg_line: str
        Portion of input line that contains only arguments

    sql_line: str
        Portion of input line that contains only SQL query/statements
    """
    arg_line, sql_line = line, ""

    # When queries include filenames, they may include SQL keywords
    #   ex. 'penguins_selected'.csv contains "select"
    # In these cases, splitting the query leads to parsing errors.
    # So we ignore any filenames by removing text between double quotes ""
    # and single quotes '' below.
    # Note: This won't affect the query because we are only modifying the
    # text we use to check for SQL commands. Any splitting is done
    # on the original line which includes filenames.
    line_no_filenames = re.sub(r"('.*')", "", line)  # 'file.csv' --> ''
    line_no_filenames = re.sub(r'(".*")', "", line_no_filenames)  # "file.csv" --> ""

    # Now that filenames are removed, check the line for any SQL commands
    # If any SQL commands are found in the line, we split the line into args and sql.
    #   Note: lines without SQL commands will not be split
    #       ex. %sql duckdb:// or %sqlplot boxplot --table data.csv
    if not any(cmd in line_no_filenames.lower() for cmd in SQL_COMMANDS):
        return arg_line, sql_line

    # Identify beginning of sql query using keywords
    split_idx = -1
    for token in line.split():
        if token.lower() in SQL_COMMANDS:
            # Found index at which to split line
            split_idx = line.find(token)
            break

    # Split line into args and sql, beginning at sql keyword
    if split_idx != -1:
        arg_line, sql_line = line[:split_idx], line[split_idx:]

    return arg_line, sql_line


def magic_args(magic_execute, line, cmd_from, allowed_duplicates=None):
    """
    Returns the parsed arguments from the line as parsed by magic_execute
    """
    allowed_duplicates = allowed_duplicates or []
    line = without_sql_comment(parser=magic_execute.parser, line=line)
    arg_line, sql_line = split_args_and_sql(line)

    args = shlex.split(arg_line, posix=False)

    if len(args) > 1:
        check_duplicate_arguments(magic_execute, cmd_from, args, allowed_duplicates)

    parsed = magic_execute.parser.parse_args(args)

    if sql_line:
        if parsed.line != "":
            parsed.line.extend(shlex.split(sql_line, posix=False))
        else:
            parsed.line = shlex.split(sql_line, posix=False)

    return parsed


def escape_string_literals_with_colon_prefix(query):
    """
    Given a query, replaces all occurrences of ':variable' with '\:variable' and
    ":variable" with "\:variable" so that the query can be passed to sqlalchemy.text
    without the literals being interpreted as bind parameters. Also calls
    escape_string_slicing_with_colon_prefix(). It doesn't replace
    the occurrences of :variable (without quotes)
    """  # noqa

    # Define the regular expression pattern for valid Python identifiers
    identifier_pattern = r"\b[a-zA-Z_][a-zA-Z0-9_]*\b"

    double_quoted_variable_pattern = r'(?<!\\)":(' + identifier_pattern + r')(?<!\\)"'

    # Define the regular expression pattern for matching ':variable' format
    single_quoted_variable_pattern = r"(?<!\\)':(" + identifier_pattern + r")(?<!\\)\'"

    # Replace ":variable" and ':variable' with "\:variable"
    query_quoted = re.sub(double_quoted_variable_pattern, r'"\\:\1"', query)
    query_quoted = re.sub(single_quoted_variable_pattern, r"'\\:\1'", query_quoted)

    double_found = re.findall(double_quoted_variable_pattern, query)
    single_found = re.findall(single_quoted_variable_pattern, query)

    # Escape occurrences of : for string slicing
    query_quoted, _ = escape_string_slicing_notation(query_quoted)

    return query_quoted, double_found + single_found


def escape_string_slicing_notation(query):
    """
    Given a query, replaces all occurrences of 'example'[x:y] with 'example'[x\:y].
    Escaping the colon using \ ensures correct string slicing behavior rather
    than being interpreted as a bind parameter.

    Parameters
    ----------
    query: str
        query to be parsed and cleaned
    """  # noqa
    identifier_pattern = r"\b[0-9_]*\b"

    # Define the regular expression pattern for matching [x:y]
    string_slicing_pattern = r"(?<!\\):(" + identifier_pattern + r")(?<!\\)\]"

    # Replace [x:y] with [x\:y]
    query_escaped = re.sub(string_slicing_pattern, r"\\:\1]", query)

    occurences_found = re.findall(string_slicing_pattern, query)

    return query_escaped, occurences_found


def find_named_parameters(input_string):
    # Define the regular expression pattern for valid Python identifiers
    identifier_pattern = r"\b[a-zA-Z_][a-zA-Z0-9_]*\b"

    # Define the regular expression pattern for matching :variable format
    variable_pattern = r'(?<!["\'])\:(' + identifier_pattern + ")"

    # Use findall to extract all matches of :variable from the input string
    matches = re.findall(variable_pattern, input_string)

    return matches



try:
    from pgspecial.main import PGSpecial
except ModuleNotFoundError:
    PGSpecial = None

from sql import exceptions


def handle_postgres_special(conn, statement):
    """Execute a PostgreSQL special statement using PGSpecial module."""
    if not PGSpecial:
        raise exceptions.MissingPackageError("pgspecial not installed")

    pgspecial = PGSpecial()
    # TODO: support for raw psycopg2 connections
    _, cur, headers, _ = pgspecial.execute(
        conn.connection_sqlalchemy.connection.cursor(), statement
    )[0]
    return FakeResultProxy(cur, headers)


class FakeResultProxy(object):
    """A fake class that pretends to behave like the ResultProxy from
    SqlAlchemy.
    """

    def __init__(self, cursor, headers):
        if cursor is None:
            cursor = []
            headers = []
        if isinstance(cursor, list):
            self.from_list(source_list=cursor)
        else:
            self.fetchall = cursor.fetchall
            self.fetchmany = cursor.fetchmany
            self.rowcount = cursor.rowcount
        self.keys = lambda: headers
        self.returns_rows = True

    def from_list(self, source_list):
        "Simulates SQLA ResultProxy from a list."

        self.fetchall = lambda: source_list
        self.rowcount = len(source_list)

        def fetchmany(size):
            pos = 0
            while pos < len(source_list):
                yield source_list[pos : pos + size]
                pos += size

        self.fetchmany = fetchmany

    def close(self):
        pass



try:
    from jupysql_plugin.widgets import ConnectorWidget
except ModuleNotFoundError:
    ConnectorWidget = None

from ploomber_core.dependencies import requires


@requires(["jupysql-plugin", "ipywidgets"])
def connect(others):
    """
    Implementation of `%sqlcmd connect`
    """

    connectorwidget = ConnectorWidget()
    return connectorwidget



import shutil


_CONDA_INSTALLED = shutil.which("conda") is not None
_PREFER_CONDA = {"psycopg2"}


def install_command(package):
    # special case for psycopg2
    if package == "psycopg2" and not _CONDA_INSTALLED:
        package = "psycopg2-binary"

    if _CONDA_INSTALLED and package in _PREFER_CONDA:
        template = "%conda install {package} -c conda-forge --yes --quiet"
    else:
        template = "%pip install {package} --quiet"

    return template.format(package=package)



"""
Plot using the SQL backend
"""

from ploomber_core.dependencies import requires
from jinja2 import Template

from sql import exceptions, display
from sql.stats import _summary_stats
from sql.util import (
    _are_numeric_values,
    validate_mutually_exclusive_args,
    to_upper_if_snowflake_conn,
    enclose_table_with_double_quotations,
)
from sql.display import message

try:
    import matplotlib.pyplot as plt
    from matplotlib.colors import Normalize
except ModuleNotFoundError:
    plt = None
    Normalize = None

try:
    import numpy as np
except ModuleNotFoundError:
    np = None

import sql.connection
import warnings


def _whishi(conn, table, column, hival, with_=None):
    if not conn:
        conn = sql.connection.ConnectionManager.current
    template = Template(
        """
SELECT COUNT(*), MAX("{{column}}")
FROM (
    SELECT "{{column}}"
    FROM {{table}}
    WHERE "{{column}}" <= {{hival}}
) AS _whishi
"""
    )

    query = template.render(table=table, column=column, hival=hival)

    values = conn.execute(query, with_).fetchone()
    keys = ["N", "wiskhi_max"]
    return {k: float(v) for k, v in zip(keys, values)}


def _whislo(conn, table, column, loval, with_=None):
    if not conn:
        conn = sql.connection.ConnectionManager.current
    template = Template(
        """
SELECT COUNT(*), MIN("{{column}}")
FROM (
    SELECT "{{column}}"
    FROM {{table}}
    WHERE "{{column}}" >= {{loval}}
) AS _whislo
"""
    )

    query = template.render(table=table, column=column, loval=loval)

    values = conn.execute(query, with_).fetchone()
    keys = ["N", "wisklo_min"]
    return {k: float(v) for k, v in zip(keys, values)}


def _percentile(conn, table, column, pct, with_=None):
    if not conn:
        conn = sql.connection.ConnectionManager.current.connection
    template = Template(
        """
SELECT
percentile_disc({{pct}}) WITHIN GROUP (ORDER BY "{{column}}") AS pct,
FROM {{table}}
"""
    )
    query = template.render(table=table, column=column, pct=pct)

    values = conn.execute(query, with_).fetchone()[0]
    return values


def _between(conn, table, column, whislo, whishi, with_=None):
    template = Template(
        """
SELECT "{{column}}"
FROM {{table}}
WHERE "{{column}}" < {{whislo}}
OR  "{{column}}" > {{whishi}}
"""
    )
    query = template.render(table=table, column=column, whislo=whislo, whishi=whishi)

    results = [float(n[0]) for n in conn.execute(query, with_).fetchall()]
    return results


# https://github.com/matplotlib/matplotlib/blob/b5ac96a8980fdb9e59c9fb649e0714d776e26701/lib/matplotlib/cbook/__init__.py
def _boxplot_stats(conn, table, column, whis=1.5, autorange=False, with_=None):
    """Compute statistics required to create a boxplot"""
    if not conn:
        conn = sql.connection.ConnectionManager.current

    def _compute_conf_interval(N, med, iqr):
        notch_min = med - 1.57 * iqr / np.sqrt(N)
        notch_max = med + 1.57 * iqr / np.sqrt(N)

        return notch_min, notch_max

    stats = dict()

    # arithmetic mean
    s_stats = _summary_stats(conn, table, column, with_=with_)

    stats["mean"] = s_stats["mean"]
    q1, med, q3 = s_stats["q1"], s_stats["med"], s_stats["q3"]
    N = s_stats["N"]

    # interquartile range
    stats["iqr"] = q3 - q1

    if stats["iqr"] == 0 and autorange:
        whis = (0, 100)

    # conf. interval around median
    stats["cilo"], stats["cihi"] = _compute_conf_interval(N, med, stats["iqr"])

    # lowest/highest non-outliers
    if np.iterable(whis) and not isinstance(whis, str):
        loval, hival = _percentile(conn, table, column, whis, with_=with_)

    elif np.isreal(whis):
        loval = q1 - whis * stats["iqr"]
        hival = q3 + whis * stats["iqr"]
    else:
        raise ValueError("whis must be a float or list of percentiles")

    # get high extreme
    wiskhi_d = _whishi(conn, table, column, hival, with_=with_)

    if wiskhi_d["N"] == 0 or wiskhi_d["wiskhi_max"] < q3:
        stats["whishi"] = q3
    else:
        stats["whishi"] = wiskhi_d["wiskhi_max"]

    # get low extreme
    wisklo_d = _whislo(conn, table, column, loval, with_=with_)

    if wisklo_d["N"] == 0 or wisklo_d["wisklo_min"] > q1:
        stats["whislo"] = q1
    else:
        stats["whislo"] = wisklo_d["wisklo_min"]

    # compute a single array of outliers
    stats["fliers"] = np.array(
        _between(conn, table, column, stats["whislo"], stats["whishi"], with_=with_)
    )

    # add in the remaining stats
    stats["q1"], stats["med"], stats["q3"] = q1, med, q3

    bxpstats = {k: v for k, v in stats.items()}

    return bxpstats


# https://github.com/matplotlib/matplotlib/blob/ddc260ce5a53958839c244c0ef0565160aeec174/lib/matplotlib/axes/_axes.py#L3915
@requires(["matplotlib"])
def boxplot(table, column, *, orient="v", with_=None, conn=None, ax=None, schema=None):
    """Plot boxplot

    Parameters
    ----------
    table : str
        Table name where the data is located

    column : str, list
        Column(s) to plot

    orient : str {"h", "v"}, default="v"
        Boxplot orientation (vertical/horizontal)

    conn : connection, default=None
        Database connection. If None, it uses the current connection

    Notes
    -----
    .. versionchanged:: 0.5.2
        Added ``with_``, and ``orient`` arguments. Added plot title and axis labels.
        Allowing to pass lists in ``column``. Function returns a ``matplotlib.Axes``
        object.

    .. versionadded:: 0.4.4

    Returns
    -------
    ax : matplotlib.Axes
        Generated plot

    Examples
    --------
    .. plot:: ../examples/plot_boxplot.py

    **Customize plot:**

    .. plot:: ../examples/plot_boxplot_custom.py

    **Horizontal boxplot:**

    .. plot:: ../examples/plot_boxplot_horizontal.py

    **Plot multiple columns from the same table:**

    .. plot:: ../examples/plot_boxplot_many.py
    """
    if not conn:
        conn = sql.connection.ConnectionManager.current

    _table = enclose_table_with_double_quotations(table, conn)
    if schema:
        _table = f'"{schema}"."{_table}"'

    ax = ax or plt.gca()
    vert = orient == "v"

    set_ticklabels = ax.set_xticklabels if vert else ax.set_yticklabels
    set_label = ax.set_ylabel if vert else ax.set_xlabel

    if isinstance(column, str):
        stats = [_boxplot_stats(conn, _table, column, with_=with_)]
        ax.bxp(stats, vert=vert)
        ax.set_title(f"{column!r} from {table!r}")
        set_label(column)
        set_ticklabels([column])
    else:
        stats = [_boxplot_stats(conn, _table, col, with_=with_) for col in column]
        ax.bxp(stats, vert=vert)
        ax.set_title(f"Boxplot from {table!r}")
        set_ticklabels(column)

    return ax


def _min_max(conn, table, column, with_=None, use_backticks=False):
    if not conn:
        conn = sql.connection.ConnectionManager.current
    template_ = """
SELECT
    MIN("{{column}}"),
    MAX("{{column}}")
FROM {{table}}
"""
    if use_backticks:
        template_ = template_.replace('"', "`")
        table = table.replace('"', "`")
    template = Template(template_)
    query = template.render(table=table, column=column)
    min_, max_ = conn.execute(query, with_).fetchone()
    return min_, max_


def _get_bar_width(ax, bins, bin_size, binwidth):
    """
    Return a single bar width based on number of bins
    or a list of bar widths if `breaks` is given.
    If bins values are str, calculate value based on figure size.

    Parameters
    ----------
    ax : matplotlib.Axes
        Generated plot

    bins : tuple
        Contains bins' midpoints as float

    bin_size : int or list or None
        Calculated bin_size from the _histogram function
        or from consecutive differences in `breaks`

    binwidth : int or float or None
        Specified binwidth from a user

    Returns
    -------
    width : float
        A single bar width
    """
    if _are_numeric_values(bin_size) or isinstance(bin_size, list):
        width = bin_size
    elif _are_numeric_values(binwidth):
        width = binwidth
    else:
        fig = plt.gcf()
        bbox = ax.get_window_extent()
        width_inch = bbox.width / fig.dpi
        width = width_inch / len(bins)

    return width


@requires(["matplotlib"])
def histogram(
    table,
    column,
    bins,
    with_=None,
    conn=None,
    category=None,
    cmap=None,
    color=None,
    edgecolor=None,
    ax=None,
    facet=None,
    breaks=None,
    binwidth=None,
    schema=None,
):
    """Plot histogram

    Parameters
    ----------
    table : str
        Table name where the data is located

    column : str, list
        Column(s) to plot

    bins : int
        Number of bins

    conn : connection, default=None
        Database connection. If None, it uses the current connection

    Notes
    -----
    .. versionchanged:: 0.5.2
        Added plot title and axis labels. Allowing to pass lists in ``column``.
        Function returns a ``matplotlib.Axes`` object.

    .. versionchanged:: 0.7.9
        Added support for NULL values, additional filter query with new logic.
        Skips the rows with NULL in the column, does not raise ValueError

    Returns
    -------
    ax : matplotlib.Axes
        Generated plot

    Examples
    --------
    .. plot:: ../examples/plot_histogram.py

    **Plot multiple columns from the same table**:

    .. plot:: ../examples/plot_histogram_many.py
    """
    if not conn:
        conn = sql.connection.ConnectionManager.current
    if isinstance(breaks, list):
        if len(breaks) < 2:
            raise exceptions.ValueError(
                f"Breaks given : {breaks}. When using breaks, please ensure "
                "to specify at least two points."
            )
        if not all([b2 > b1 for b1, b2 in zip(breaks[:-1], breaks[1:])]):
            raise exceptions.ValueError(
                f"Breaks given : {breaks}. When using breaks, please ensure that "
                "breaks are strictly increasing."
            )

    if _are_numeric_values(binwidth):
        if binwidth <= 0:
            raise exceptions.ValueError(
                f"Binwidth given : {binwidth}. When using binwidth, please ensure to "
                "pass a positive value."
            )
        binwidth = float(binwidth)
    elif binwidth is not None:
        raise exceptions.ValueError(
            f"Binwidth given : {binwidth}. When using binwidth, please ensure to "
            "pass a numeric value."
        )

    validate_mutually_exclusive_args(
        ["bins", "breaks", "binwidth"], [bins, breaks, binwidth]
    )

    _table = enclose_table_with_double_quotations(table, conn)
    if schema:
        _table = f'"{schema}"."{_table}"'

    ax = ax or plt.gca()

    if category:
        if isinstance(column, list):
            if len(column) > 1:
                raise ValueError(
                    f"""Columns given : {column}.
                    When using a stacked histogram,
                    please ensure that you specify only one column."""
                )
            else:
                column = " ".join(column)

        if column is None or len(column) == 0:
            raise ValueError("Column name has not been specified")

        bin_, height, bin_size = _histogram(
            _table,
            column,
            bins,
            with_=with_,
            conn=conn,
            breaks=breaks,
            binwidth=binwidth,
        )
        width = _get_bar_width(ax, bin_, bin_size, binwidth)
        data = _histogram_stacked(
            _table,
            column,
            category,
            bin_,
            bin_size,
            with_=with_,
            conn=conn,
            facet=facet,
            breaks=breaks,
            binwidth=binwidth,
        )
        cmap = plt.get_cmap(cmap or "viridis")
        norm = Normalize(vmin=0, vmax=len(data))

        bottom = np.zeros(len(bin_))
        for i, values in enumerate(data):
            values_ = values[1:]

            if isinstance(color, list):
                color_ = color[0]
                if len(color) > 1:
                    warnings.warn(
                        "If you want to colorize each bar with multiple "
                        "colors please use cmap attribute instead "
                        "of 'fill'",
                        UserWarning,
                    )
            else:
                color_ = color or cmap(norm(i + 1))

            if isinstance(edgecolor, list):
                edgecolor_ = edgecolor[0]
            else:
                edgecolor_ = edgecolor or "None"

            ax.bar(
                bin_,
                values_,
                align="center",
                label=values[0],
                width=width,
                bottom=bottom,
                edgecolor=edgecolor_,
                color=color_,
            )
            bottom += values_

        ax.set_title(f"Histogram from {table!r}")
        # reverses legend order so alphabetically first goes on top
        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles[::-1], labels[::-1])
    elif isinstance(column, str):
        bin_, height, bin_size = _histogram(
            _table,
            column,
            bins,
            with_=with_,
            conn=conn,
            facet=facet,
            breaks=breaks,
            binwidth=binwidth,
        )
        width = _get_bar_width(ax, bin_, bin_size, binwidth)

        ax.bar(
            bin_,
            height,
            align="center",
            width=width,
            color=color,
            edgecolor=edgecolor or "None",
            label=column,
        )
        ax.set_title(f"{column!r} from {table!r}")
        ax.set_xlabel(column)

    else:
        if breaks and len(column) > 1:
            raise exceptions.UsageError(
                "Multiple columns don't support breaks. Please use bins instead."
            )
        for i, col in enumerate(column):
            bin_, height, bin_size = _histogram(
                _table,
                col,
                bins,
                with_=with_,
                conn=conn,
                facet=facet,
                breaks=breaks,
                binwidth=binwidth,
            )
            width = _get_bar_width(ax, bin_, bin_size, binwidth)

            if isinstance(color, list):
                color_ = color[i]
            else:
                color_ = color

            if isinstance(edgecolor, list):
                edgecolor_ = edgecolor[i]
            else:
                edgecolor_ = edgecolor or "None"

            ax.bar(
                bin_,
                height,
                align="center",
                width=width,
                alpha=0.5,
                label=col,
                color=color_,
                edgecolor=edgecolor_,
            )
            ax.set_title(f"Histogram from {table!r}")
            ax.legend()

    ax.set_ylabel("Count")

    return ax


def _histogram(
    table, column, bins, with_=None, conn=None, facet=None, breaks=None, binwidth=None
):
    """Compute bins and heights"""
    if not conn:
        conn = sql.connection.ConnectionManager.current
    use_backticks = conn.is_use_backtick_template()

    # Snowflake will use UPPERCASE in the table and column name
    column = to_upper_if_snowflake_conn(conn, column)
    table = to_upper_if_snowflake_conn(conn, table)
    # FIXME: we're computing all the with elements twice
    min_, max_ = _min_max(conn, table, column, with_=with_, use_backticks=use_backticks)

    # Define all relevant filters here
    filter_query_1 = f'"{column}" IS NOT NULL'

    filter_query_2 = f"{facet['key']} == '{facet['value']}'" if facet else None

    filter_query = _filter_aggregate(filter_query_1, filter_query_2)

    bin_size = None

    if _are_numeric_values(min_, max_):
        if breaks:
            if min_ > breaks[-1]:
                raise exceptions.UsageError(
                    f"All break points are lower than the min data point of {min_}."
                )
            elif max_ < breaks[0]:
                raise exceptions.UsageError(
                    f"All break points are higher than the max data point of {max_}."
                )

            cases, bin_size = [], []
            for b_start, b_end in zip(breaks[:-1], breaks[1:]):
                case = f"WHEN {{{{column}}}} > {b_start} AND {{{{column}}}} <= {b_end} \
                        THEN {(b_start+b_end)/2}"
                cases.append(case)
                bin_size.append(b_end - b_start)
            cases[0] = cases[0].replace(">", ">=", 1)
            bin_midpoints = [
                (b_start + b_end) / 2 for b_start, b_end in zip(breaks[:-1], breaks[1:])
            ]
            all_bins = " union ".join([f"select {mid} as bin" for mid in bin_midpoints])

            # Group data based on the intervals in breaks
            # Left join is used to ensure count=0
            template_ = (
                "select all_bins.bin, coalesce(count_table.count, 0) as count "
                f"from ({all_bins}) as all_bins "
                "left join ("
                f"select case {' '.join(cases)} end as bin, "
                "count(*) as count "
                "from {{table}} "
                "{{filter_query}} "
                "group by bin) "
                "as count_table on all_bins.bin = count_table.bin "
                "order by all_bins.bin;"
            )

            breaks_filter_query = (
                f'"{column}" >= {breaks[0]} and "{column}" <= {breaks[-1]}'
            )
            filter_query = _filter_aggregate(
                filter_query_1, filter_query_2, breaks_filter_query
            )

            if use_backticks:
                template_ = template_.replace('"', "`")
                table = table.replace('"', "`")

            template = Template(template_)

            query = template.render(
                table=table, column=column, filter_query=filter_query
            )
        elif not binwidth and not isinstance(bins, int):
            raise ValueError(
                f"bins are '{bins}'. Please specify a valid number of bins."
            )
        else:
            # Use bins - 1 instead of bins and round half down instead of floor
            # to mimic right-closed histogram intervals in R ggplot
            range_ = max_ - min_
            if binwidth:
                bin_size = binwidth
                if binwidth > range_:
                    message(
                        f"Specified binwidth {binwidth} is larger than "
                        f"the range {range_}. Please choose a smaller binwidth."
                    )
            else:
                bin_size = range_ / (bins - 1)
            template_ = """
                select
                ceiling("{{column}}"/{{bin_size}} - 0.5)*{{bin_size}} as bin,
                count(*) as count
                from {{table}}
                {{filter_query}}
                group by bin
                order by bin;
                """

            if use_backticks:
                template_ = template_.replace('"', "`")
                table = table.replace('"', "`")

            template = Template(template_)

            query = template.render(
                table=table, column=column, bin_size=bin_size, filter_query=filter_query
            )
    else:
        template_ = """
        select
            "{{column}}" as col, count ("{{column}}")
        from {{table}}
        {{filter_query}}
        group by col
        order by col;
        """

        if use_backticks:
            template_ = template_.replace('"', "`")
            table = table.replace('"', "`")

        template = Template(template_)

        query = template.render(table=table, column=column, filter_query=filter_query)

    data = conn.execute(query, with_).fetchall()

    bin_, height = zip(*data)

    return bin_, height, bin_size


def _histogram_stacked(
    table,
    column,
    category,
    bins,
    bin_size,
    with_=None,
    conn=None,
    facet=None,
    breaks=None,
    binwidth=None,
):
    """Compute the corresponding heights of each bin based on the category"""
    if not conn:
        conn = sql.connection.ConnectionManager.current

    cases = []
    if breaks:
        breaks_filter_query = (
            f'"{column}" >= {breaks[0]} and "{column}" <= {breaks[-1]}'
        )
        for b_start, b_end in zip(breaks[:-1], breaks[1:]):
            case = f'SUM(CASE WHEN {column} > {b_start} AND {column} <= {b_end} \
                    THEN 1 ELSE 0 END) AS "{(b_start+b_end)/2}",'
            cases.append(case)
        cases[0] = cases[0].replace(">", ">=", 1)
    else:
        if binwidth:
            bin_size = binwidth
        tolerance = bin_size / 1000  # Use to avoid floating point error
        for bin in bins:
            # Use round half down instead of floor to mimic
            # right-closed histogram intervals in R ggplot
            case = (
                f"SUM(CASE WHEN ABS(CEILING({column}/{bin_size} - 0.5)*{bin_size} "
                f"- {bin}) <= {tolerance} THEN 1 ELSE 0 END) AS '{bin}',"
            )
            cases.append(case)

    cases = " ".join(cases)

    filter_query_1 = f'"{column}" IS NOT NULL'

    filter_query_2 = f"{facet['key']} == '{facet['value']}'" if facet else None

    if breaks:
        filter_query = _filter_aggregate(
            filter_query_1, filter_query_2, breaks_filter_query
        )
    else:
        filter_query = _filter_aggregate(filter_query_1, filter_query_2)

    template = Template(
        """
        SELECT {{category}},
        {{cases}}
        FROM {{table}}
        {{filter_query}}
        GROUP BY {{category}}
        ORDER BY {{category}} DESC;
        """
    )
    query = template.render(
        table=table,
        column=column,
        bin_size=bin_size,
        category=category,
        filter_query=filter_query,
        cases=cases,
    )

    data = conn.execute(query, with_).fetchall()

    return data


def _filter_aggregate(*filter_queries):
    """Return a single filter query based on multiple queries.

    Parameters:
    ----------
    *filter_queries (str):
    Variable length argument list of filter queries.
    Filter query is  string with a filtering condition in SQL
    (e.g., "age > 25").
    (e.g., "column is NULL").

    Notes
    -----
    .. versionadded:: 0.7.9

    Returns:
    -----
    final_filter (str):
    A string that represents a SQL WHERE clause

    """
    final_filter = ""
    for idx, query in enumerate(filter_queries):
        if query is not None:
            if idx == 0:
                final_filter = f"{final_filter}WHERE {query}"
                continue
            final_filter = f"{final_filter} AND {query}"
    return final_filter


def _bar(table, column, with_=None, conn=None):
    """get x and height for bar plot"""
    if not conn:
        conn = sql.connection.ConnectionManager.current
    use_backticks = conn.is_use_backtick_template()

    if isinstance(column, list):
        if len(column) > 2:
            raise exceptions.UsageError(
                f"Passed columns: {column}\n"
                "Bar chart currently supports, either a single column"
                " on which group by and count is applied or"
                " two columns: labels and size"
            )

        x_ = column[0]
        height_ = column[1]

        display.message(f"Removing NULLs, if there exists any from {x_} and {height_}")
        template_ = """
            select "{{x_}}" as x,
            "{{height_}}" as height
            from {{table}}
            where "{{x_}}" is not null
            and "{{height_}}" is not null;
            """

        xlabel = x_
        ylabel = height_

        if use_backticks:
            template_ = template_.replace('"', "`")
            table = table.replace('"', "`")

        template = Template(template_)
        query = template.render(table=table, x_=x_, height_=height_)

    else:
        display.message(f"Removing NULLs, if there exists any from {column}")
        template_ = """
                select "{{column}}" as x,
                count("{{column}}") as height
                from {{table}}
                where "{{column}}" is not null
                group by "{{column}}";
                """

        xlabel = column
        ylabel = "Count"

        if use_backticks:
            template_ = template_.replace('"', "`")
            table = table.replace('"', "`")

        template = Template(template_)
        query = template.render(table=table, column=column)

    data = conn.execute(query, with_).fetchall()

    x, height = zip(*data)

    if x[0] is None:
        raise ValueError("Data contains NULLs")

    return x, height, xlabel, ylabel


@requires(["matplotlib"])
def bar(
    table,
    column,
    show_num=False,
    orient="v",
    with_=None,
    conn=None,
    cmap=None,
    color=None,
    edgecolor=None,
    ax=None,
    schema=None,
):
    """Plot Bar Chart

    Parameters
    ----------
    table : str
        Table name where the data is located

    column : str
        Column(s) to plot

    show_num: bool
        Show numbers on top of plot

    orient : str, default='v'
        Orientation of the plot. 'v' for vertical and 'h' for horizontal

    conn : connection, default=None
        Database connection. If None, it uses the current connection

    Notes
    -----

    .. versionadded:: 0.7.6

    Returns
    -------
    ax : matplotlib.Axes
        Generated plot

    """

    if not conn:
        conn = sql.connection.ConnectionManager.current

    _table = enclose_table_with_double_quotations(table, conn)
    if schema:
        _table = f'"{schema}"."{_table}"'

    ax = ax or plt.gca()

    if column is None:
        raise exceptions.UsageError("Column name has not been specified")

    x, height_, xlabel, ylabel = _bar(_table, column, with_=with_, conn=conn)

    if color and cmap:
        # raise a userwarning
        warnings.warn(
            "Both color and cmap are given. cmap will be ignored", UserWarning
        )

    if (not color) and cmap:
        cmap = plt.get_cmap(cmap)
        norm = Normalize(vmin=0, vmax=len(x))
        color = [cmap(norm(i)) for i in range(len(x))]

    if orient == "h":
        ax.barh(
            x,
            height_,
            align="center",
            edgecolor=edgecolor,
            color=color,
        )
        ax.set_xlabel(ylabel)
        ax.set_ylabel(xlabel)
    else:
        ax.bar(
            x,
            height_,
            align="center",
            edgecolor=edgecolor,
            color=color,
        )
        ax.set_ylabel(ylabel)
        ax.set_xlabel(xlabel)

    if show_num:
        if orient == "v":
            for i, v in enumerate(height_):
                ax.text(
                    i,
                    v,
                    str(v),
                    color="black",
                    fontweight="bold",
                    ha="center",
                    va="bottom",
                )
        else:
            for i, v in enumerate(height_):
                ax.text(
                    v,
                    i,
                    str(v),
                    color="black",
                    fontweight="bold",
                    ha="left",
                    va="center",
                )

    ax.set_title(table)

    return ax


def _pie(table, column, with_=None, conn=None):
    """get x and height for pie chart"""
    if not conn:
        conn = sql.connection.ConnectionManager.current
    use_backticks = conn.is_use_backtick_template()

    if isinstance(column, list):
        if len(column) > 2:
            raise exceptions.UsageError(
                f"Passed columns: {column}\n"
                "Pie chart currently supports, either a single column"
                " on which group by and count is applied or"
                " two columns: labels and size"
            )

        labels_ = column[0]
        size_ = column[1]

        display.message(
            f"Removing NULLs, if there exists any from {labels_} and {size_}"
        )
        template_ = """
                select "{{labels_}}" as labels,
                "{{size_}}" as size
                from {{table}}
                where "{{labels_}}" is not null
                and "{{size_}}" is not null;
                """
        if use_backticks:
            template_ = template_.replace('"', "`")
            table = table.replace('"', "`")

        template = Template(template_)
        query = template.render(table=table, labels_=labels_, size_=size_)

    else:
        display.message(f"Removing NULLs, if there exists any from {column}")
        template_ = """
                select "{{column}}" as x,
                count("{{column}}") as height
                from {{table}}
                where "{{column}}" is not null
                group by "{{column}}";
                """
        if use_backticks:
            template_ = template_.replace('"', "`")
            table = table.replace('"', "`")

        template = Template(template_)
        query = template.render(table=table, column=column)

    data = conn.execute(query, with_).fetchall()

    labels, size = zip(*data)

    if labels[0] is None:
        raise ValueError("Data contains NULLs")

    return labels, size


@requires(["matplotlib"])
def pie(
    table,
    column,
    show_num=False,
    with_=None,
    conn=None,
    cmap=None,
    color=None,
    ax=None,
    schema=None,
):
    """Plot Pie Chart

    Parameters
    ----------
    table : str
        Table name where the data is located

    column : str
        Column(s) to plot

    show_num: bool
        Show numbers on top of plot

    conn : connection, default=None
        Database connection. If None, it uses the current connection

    Notes
    -----

    .. versionadded:: 0.7.6

    Returns
    -------
    ax : matplotlib.Axes
        Generated plot
    """

    if not conn:
        conn = sql.connection.ConnectionManager.current

    _table = enclose_table_with_double_quotations(table, conn)
    if schema:
        _table = f'"{schema}"."{_table}"'

    ax = ax or plt.gca()

    if column is None:
        raise exceptions.UsageError("Column name has not been specified")

    labels, size_ = _pie(_table, column, with_=with_, conn=conn)

    if color and cmap:
        # raise a userwarning
        warnings.warn(
            "Both color and cmap are given. cmap will be ignored", UserWarning
        )

    if (not color) and cmap:
        cmap = plt.get_cmap(cmap)
        norm = Normalize(vmin=0, vmax=len(labels))
        color = [cmap(norm(i)) for i in range(len(labels))]

    if show_num:
        ax.pie(
            size_,
            labels=labels,
            colors=color,
            autopct="%1.2f%%",
        )
    else:
        ax.pie(
            size_,
            labels=labels,
            colors=color,
        )

    ax.set_title(table)

    return ax



"""
In most scenarios, users don't care about the full Python traceback because it's
irrelevant to them (they run SQL, not Python code). Hence, when raising errors,
we only display the error message. This is possible via IPython.core.error.UsageError:
IPython/Jupyter automatically detect this error and hide the traceback.
Unfortunately, IPython.core.error.UsageError isn't the most appropriate error type for
all scenarios, so we define our own error types here. The main caveat is that due to a
bug in IPython (https://github.com/ipython/ipython/issues/14024), subclassing
IPython.core.error.UsageError doesn't work, so `exception_factory` is a workaround
to create new errors that are IPython.core.error.UsageError but with a different name.

"""

from IPython.core import error


def exception_factory(name):
    def _error(message):
        exc = error.UsageError(message)
        exc.error_type = name
        return exc

    return _error


# raise it when there's an issue with the user's input in a magic. e.g., missing an
# argument
UsageError = exception_factory("UsageError")

# raise it when a user wants to use a feature that requires an optional dependency
MissingPackageError = exception_factory("MissingPackageError")

# the following exceptions should be called instead of the Python built-in ones,
# for guidelines on when to use them:
# https://docs.python.org/3/library/exceptions.html#bltin-exceptions
TypeError = exception_factory("TypeError")
RuntimeError = exception_factory("RuntimeError")
ValueError = exception_factory("ValueError")
KeyError = exception_factory("KeyError")
FileNotFoundError = exception_factory("FileNotFoundError")
NotImplementedError = exception_factory("NotImplementedError")

# The following are internal exceptions that should not be raised directly

# raised internally when the user chooses a table that doesn't exist
TableNotFoundError = exception_factory("TableNotFoundError")

# raise it when there is an error in parsing the configuration file
ConfigurationError = exception_factory("ConfigurationError")


InvalidQueryParameters = exception_factory("InvalidQueryParameters")



from sql.widgets import TableWidget
from sql.cmd.cmd_utils import CmdParser
from sql.util import expand_args, is_rendering_required


def explore(others, user_ns):
    """
    Implementation of `%sqlcmd explore`
    This function takes in a string containing command line arguments,
    parses them to extract the name of the table, and displays an interactive
    widget for exploring the contents of the specified table. It also uses the
    kernel namespace for expanding arguments declared as variables.

    Parameters
    ----------
    others : str,
        A string containing the command line arguments.

    user_ns : dict,
        User namespace of IPython kernel

    """
    parser = CmdParser()
    parser.add_argument("-t", "--table", type=str, help="Table name", required=True)
    parser.add_argument("-s", "--schema", type=str, help="Schema name", required=False)
    args = parser.parse_args(others)
    if is_rendering_required(" ".join(others)):
        expand_args(args, user_ns)

    table_widget = TableWidget(args.table, args.schema)
    return table_widget



import warnings
import difflib
from sql import exceptions, display
import json
from pathlib import Path
from sqlglot import parse_one, exp
from sqlglot.errors import ParseError
from sqlalchemy.exc import SQLAlchemyError
from ploomber_core.dependencies import requires

try:
    from pyspark.sql.utils import AnalysisException
except ModuleNotFoundError:
    AnalysisException = None

import ast
from os.path import isfile
import re

from jinja2 import Template


try:
    import toml
except ModuleNotFoundError:
    toml = None

SINGLE_QUOTE = "'"
DOUBLE_QUOTE = '"'

CONFIGURATION_DOCS_STR = "https://jupysql.ploomber.io/en/latest/api/configuration.html#loading-from-a-file"  # noqa


def sanitize_identifier(identifier):
    if (identifier[0] == SINGLE_QUOTE and identifier[-1] == SINGLE_QUOTE) or (
        identifier[0] == DOUBLE_QUOTE and identifier[-1] == DOUBLE_QUOTE
    ):
        return identifier[1:-1]
    else:
        return identifier


def convert_to_scientific(value):
    """
    Converts value to scientific notation if necessary

    Parameters
    ----------
    value : any
        Value to format.
    """
    if (
        isinstance(value, (int, float))
        and not isinstance(value, bool)
        and _is_long_number(value)
    ):
        new_value = "{:,.3e}".format(value)

    else:
        new_value = value

    return new_value


def _is_long_number(num) -> bool:
    """
    Checks if num's digits > 10
    """
    if "." in str(num):
        split_by_decimal = str(num).split(".")
        if len(split_by_decimal[0]) > 10 or len(split_by_decimal[1]) > 10:
            return True
    return False


def get_suggestions_message(suggestions):
    suggestions_message = ""
    if len(suggestions) > 0:
        _suggestions_string = pretty_print(suggestions, last_delimiter="or")
        suggestions_message = f"\nDid you mean: {_suggestions_string}"
    return suggestions_message


def pretty_print(
    obj: list, delimiter: str = ",", last_delimiter: str = "and", repr_: bool = False
) -> str:
    """
    Returns a formatted string representation of an array
    """
    if repr_:
        sorted_ = sorted(repr(element) for element in obj)
    else:
        sorted_ = sorted(f"'{element}'" for element in obj)

    if len(sorted_) > 1:
        sorted_[-1] = f"{last_delimiter} {sorted_[-1]}"

    return f"{delimiter} ".join(sorted_)


def strip_multiple_chars(string: str, chars: str) -> str:
    """
    Trims characters from the start and end of the string
    """
    return string.translate(str.maketrans("", "", chars))


def flatten(src, ltypes=(list, tuple)):
    """The flatten function creates a new tuple / list
    with all sub-tuple / sub-list elements concatenated into it recursively

    Parameters
    ----------
    src : tuple / list
        Source tuple / list with all sub-tuple / sub-list elements
    ltypes : tuple, optional
        sub element's data type, by default (list, tuple)

    Returns
    -------
    tuple / list
        Flatten tuple / list
    """
    ltype = type(src)
    # Create a process list to handle flatten elements
    process_list = list(src)
    i = 0
    while i < len(process_list):
        while isinstance(process_list[i], ltypes):
            if not process_list[i]:
                process_list.pop(i)
                i -= 1
                break
            else:
                process_list[i : i + 1] = process_list[i]
        i += 1

    # If input src data type is tuple, return tuple
    if not isinstance(process_list, ltype):
        return tuple(process_list)
    return process_list


def parse_sql_results_to_json(rows, columns) -> str:
    """
    Serializes sql rows to a JSON formatted ``str``
    """
    dicts = [dict(zip(list(columns), row)) for row in rows]
    rows_json = json.dumps(dicts, indent=4, sort_keys=True, default=str).replace(
        "null", '"None"'
    )

    return rows_json


def show_deprecation_warning():
    """
    Raises CTE deprecation warning
    """
    warnings.warn(
        "CTE dependencies are now automatically inferred, "
        "you can omit the --with arguments. Using --with will "
        "raise an exception in the next major release so please remove it.",
        FutureWarning,
    )


def check_duplicate_arguments(
    magic_execute, cmd_from, args, allowed_duplicates=None, disallowed_aliases=None
) -> bool:
    """
    Raises UsageError when duplicate arguments are passed to magics.
    Returns true if no duplicates in arguments or aliases.

    Parameters
    ----------
    magic_execute
        The execute method of the magic class.
    cmd_from
        Which magic class invoked this function. One of 'sql', 'sqlplot' or 'sqlcmd'.
    args
        The arguments passed to the magic command.
    allowed_duplicates
        The duplicate arguments that are allowed for the class which invoked this
        function. Defaults to None.
    disallowed_aliases
        The aliases for the arguments that are not allowed to be used together
        for the class that invokes this function. Defaults to None.

    Returns
    -------
    boolean
        When there are no duplicates, a True bool is returned.
    """
    allowed_duplicates = allowed_duplicates or []
    disallowed_aliases = disallowed_aliases or {}

    aliased_arguments = {}
    unaliased_arguments = []

    # Separates the aliased_arguments and unaliased_arguments.
    # Aliased arguments example: '-w' and '--with'
    if cmd_from != "sqlcmd":
        for decorator in magic_execute.decorators:
            decorator_args = decorator.args
            if len(decorator_args) > 1:
                aliased_arguments[decorator_args[0]] = decorator_args[1]
            else:
                if decorator_args[0].startswith("--") or decorator_args[0].startswith(
                    "-"
                ):
                    unaliased_arguments.append(decorator_args[0])

    if aliased_arguments == {}:
        aliased_arguments = disallowed_aliases

    # Separate arguments from passed options
    args = [arg for arg in args if arg.startswith("--") or arg.startswith("-")]

    # Separate single and double hyphen arguments
    # Using sets here for better performance of looking up hash tables
    single_hyphen_opts = set()
    double_hyphen_opts = set()

    for arg in args:
        if arg.startswith("--"):
            double_hyphen_opts.add(arg)
        elif arg.startswith("-"):
            single_hyphen_opts.add(arg)

    # Get duplicate arguments
    duplicate_args = []
    visited_args = set()
    for arg in args:
        if arg not in allowed_duplicates:
            if arg not in visited_args:
                visited_args.add(arg)
            else:
                duplicate_args.append(arg)

    # Check if alias pairs are present and track the pair for the error message
    # Example: would filter out `-w` and `--with` if both are present
    alias_pairs_present = [
        (opt, aliased_arguments[opt])
        for opt in single_hyphen_opts
        if opt in aliased_arguments
        if aliased_arguments[opt] in double_hyphen_opts
    ]

    # Generate error message based on presence of duplicates and
    # aliased arguments
    error_message = ""
    if duplicate_args:
        duplicates_error = (
            f"Duplicate arguments in %{cmd_from}. "
            "Please use only one of each of the following: "
            f"{', '.join(sorted(duplicate_args))}. "
        )
    else:
        duplicates_error = ""

    if alias_pairs_present:
        arg_list = sorted([" or ".join(pair) for pair in alias_pairs_present])
        alias_error = (
            f"Duplicate aliases for arguments in %{cmd_from}. "
            "Please use either one of "
            f"{', '.join(arg_list)}."
        )
    else:
        alias_error = ""

    error_message = f"{duplicates_error}{alias_error}"

    # If there is an error message to be raised, raise it
    if error_message:
        raise exceptions.UsageError(error_message)

    return True


def find_path_from_root(file_name):
    """
    Recursively finds an absolute path to file_name starting
    from current to root directory
    """
    current = Path().resolve()
    while not (current / file_name).exists():
        if current == current.parent:
            return None

        current = current.parent

    return Path(current, file_name)


def find_close_match(word, possibilities):
    """Find closest match between invalid input and possible options"""
    return difflib.get_close_matches(word, possibilities)


def find_close_match_config(word, possibilities, n=3):
    """Finds closest matching configurations and displays message"""
    closest_matches = difflib.get_close_matches(word, possibilities, n=n)
    if not closest_matches:
        display.message_html(
            f"'{word}' is an invalid configuration. Please review our "
            "<a href='https://jupysql.ploomber.io/en/latest/api/configuration.html#options'>"  # noqa
            "configuration guideline</a>."
        )
    else:
        display.message(
            f"'{word}' is an invalid configuration. Did you mean "
            f"{pretty_print(closest_matches, last_delimiter='or')}?"
        )


def get_line_content_from_toml(file_path, line_number):
    """
    Locates a line that error occurs when loading a toml file
    and returns the line, key, and value
    """
    with open(file_path, "r") as file:
        lines = file.readlines()
        eline = lines[line_number - 1].strip()
        ekey, evalue = None, None
        if "=" in eline:
            ekey, evalue = map(str.strip, eline.split("="))
        return eline, ekey, evalue


def to_upper_if_snowflake_conn(conn, upper):
    return (
        upper.upper()
        if callable(conn._get_sqlglot_dialect)
        and conn._get_sqlglot_dialect() == "snowflake"
        else upper
    )


@requires(["toml"])
def load_toml(file_path):
    """
    Returns toml file content in a dictionary format
    and raises error if it fails to load the toml file
    """
    try:
        with open(file_path, "r") as file:
            content = file.read()
            return toml.loads(content)
    except toml.TomlDecodeError as e:
        raise parse_toml_error(e, file_path)


def parse_toml_error(e, file_path):
    eline, ekey, evalue = get_line_content_from_toml(file_path, e.lineno)
    if "Duplicate keys!" in str(e):
        return exceptions.ConfigurationError(
            f"Duplicate key found: '{ekey}' in {file_path}"
        )
    elif "Only all lowercase booleans" in str(e):
        return exceptions.ConfigurationError(
            f"Invalid value '{evalue}' in '{eline}' in {file_path}. "
            "Valid boolean values: true, false"
        )
    elif "invalid literal for int()" in str(e):
        return exceptions.ConfigurationError(
            f"Invalid value '{evalue}' in '{eline}' in {file_path}. "
            "To use str value, enclose it with ' or \"."
        )
    else:
        return e


def get_user_configs(primary_path, alternate_path):
    """
    Returns saved configuration settings in a toml file from given file_path

    Parameters
    ----------
    primary_path : Path
        file path to toml in project directory
    alternate_path : Path
        file path to ~/.jupysql/config

    Returns
    -------
    dict
        saved configuration settings
    Path
        the path of the file used to get user configurations
    """
    data = None
    display_tip = True  # Set to true if tip is to be displayed
    configuration_docs_displayed = False  # To disable showing guidelines once shown

    # Look for user configurations in pyproject.toml and ~/.jupysql/config
    # in that particular order
    path_list = [primary_path, alternate_path]
    for file_path in path_list:
        section_found = False
        if file_path and file_path.exists():
            data = load_toml(file_path)

            data = data.get("tool")

            # Look for jupysql section under tool
            if data:
                keys = data.keys()
                data = data.get("jupysql")
                if data is None:
                    similar_key = case_insensitive_match("jupysql", keys)
                    if similar_key:
                        display.message(
                            f"Hint: We found 'tool.{similar_key}' in {file_path}. "
                            f"Did you mean 'tool.jupysql'?"
                        )

            # Look for SqlMagic section under jupysql
            if data:
                keys = data.keys()
                data = data.get("SqlMagic")
                if data is None:
                    similar_key_list = find_close_match("SqlMagic", keys)
                    if similar_key_list:
                        raise exceptions.ConfigurationError(
                            f"[tool.jupysql.{similar_key_list[0]}] is an "
                            f"invalid section name in {file_path}. "
                            f"Did you mean [tool.jupysql.SqlMagic]?"
                        )

            if data is None:
                if display_tip:
                    display.message(
                        f"Tip: You may define configurations in {primary_path}"
                        f" or {alternate_path}. "
                    )
                    display_tip = False
            elif data == {}:
                section_found = True
                display.message(
                    f"[tool.jupysql.SqlMagic] present in {file_path} but empty. "
                )
                display_tip = False
            else:
                section_found = True

        if not display_tip and not configuration_docs_displayed:
            display.message_html(
                f"Please review our <a href='{CONFIGURATION_DOCS_STR}'>"
                "configuration guideline</a>."
            )
            configuration_docs_displayed = True

        if not data and not section_found and file_path and file_path.exists():
            display.message(f"Did not find user configurations in {file_path}.")
        elif section_found and data:
            return data, file_path

    return None, None


def get_default_configs(sql):
    """
    Returns a dictionary of SqlMagic configuration settings users can set
    with their default values.
    """
    default_configs = sql.trait_defaults()
    del default_configs["parent"]
    del default_configs["config"]
    return default_configs


def _are_numeric_values(*values):
    return all([isinstance(value, (int, float)) for value in values])


def validate_mutually_exclusive_args(arg_names, args):
    """
    Raises ValueError if a list of values from arg_names filtered by
    args' boolean representations is longer than one.

    Parameters
    ----------
    arg_names : list
        args' names in string
    args : list
        args values
    """
    specified_args = [arg_name for arg_name, arg in zip(arg_names, args) if arg]
    if len(specified_args) > 1:
        raise exceptions.ValueError(
            f"{pretty_print(specified_args)} are specified. "
            "You can only specify one of them."
        )


def validate_nonidentifier_connection(arg):
    """
    Raises UsageError if a connection is passed to `%sql/%%sql` through
    object property, list, or dictionary.

    Parameters
    ----------
    arg : str
        argument to check whether it is a valid connection or not
    """
    if not arg.isidentifier() and is_valid_python_code(arg) and not arg.endswith(";"):
        raise exceptions.UsageError(
            f"'{arg}' is not a valid connection identifier. "
            "Please pass the variable's name directly, as passing "
            "object attributes, dictionaries or lists won't work."
        )


def is_valid_python_code(code):
    try:
        ast.parse(code)
        return True
    except SyntaxError:
        return False


def extract_tables_from_query(query):
    """
    Function to extract names of tables from
    a syntactically correct query

    Parameters
    ----------
    query : str, user query

    Returns
    -------
    list
        List of tables in the query
        [] if error in parsing the query
    """
    try:
        tables = [
            table.name
            for table in parse_one(query).find_all(exp.Table)
            if hasattr(table, "name")
        ]
        return tables
    except ParseError:
        # TODO : Instead of returning [] return the
        # exact parse error
        return []


def is_sqlalchemy_error(error):
    """Function to check if error is SQLAlchemy error"""
    return isinstance(error, SQLAlchemyError)


def is_non_sqlalchemy_error(error):
    """Function to check if error is a specific non-SQLAlchemy error"""
    specific_db_errors = [
        "duckdb.CatalogException",
        "Catalog Error",
        "Parser Error",
        "pyodbc.ProgrammingError",
        # Clickhouse errors
        "DB::Exception:",
    ]
    is_pyspark_analysis_exception = (
        isinstance(error, AnalysisException) if AnalysisException else False
    )
    return (
        any(msg in str(error) for msg in specific_db_errors)
        or is_pyspark_analysis_exception
    )


def if_substring_exists(string, substrings):
    """Function to check if any of substring in
    substrings exist in string"""
    return any((msg in string) or (re.search(msg, string)) for msg in substrings)


def enclose_table_with_double_quotations(table, conn):
    """
    Function to enclose a file path, schema name,
    or table name with double quotations
    """
    if isfile(table):
        _table = f'"{table}"'
    elif "." in table and not table.startswith('"'):
        parts = table.split(".")
        _table = f'"{parts[0]}"."{parts[1]}"'
    else:
        _table = table

    use_backticks = conn.is_use_backtick_template()
    if use_backticks:
        _table = _table.replace('"', "`")

    return _table


def is_rendering_required(line):
    """Function to check possibility of line
    text containing expandable arguments"""

    return "{{" in line and "}}" in line


def render_string_using_namespace(value, user_ns):
    """
    Function to substitute command line arguments
    with variables defined by user in the IPython
    kernel.

    Parameters
    ----------
    value : str,
        text to be rendered

    user_ns : dict,
        User namespace of IPython kernel
    """

    if isinstance(value, str) and value.startswith("{{") and value.endswith("}}"):
        return Template(value).render(user_ns)
    return value


def expand_args(args, user_ns):
    """
    Function to substitute command line arguments
    with variables defined by user in the IPython
    kernel.

    Parameters
    ----------
    args : argparse.Namespace,
        object to hold the command line arguments.

    user_ns : dict,
        User namespace of IPython kernel
    """

    for attribute in vars(args):
        value = getattr(args, attribute)
        if value:
            if isinstance(value, list):
                substituted_value = []
                for item in value:
                    rendered_value = render_string_using_namespace(item, user_ns)
                    substituted_value.append(rendered_value)
                setattr(args, attribute, substituted_value)
            else:
                rendered_value = render_string_using_namespace(value, user_ns)
                setattr(args, attribute, rendered_value)


def case_insensitive_match(target, string_list):
    """
    Perform a case-insensitive match of a target string against a list of strings.

    Parameters
    ----------
    target : str
        The target string to match.
    string_list : list of str
        The list of strings to search through.

    Returns
    -------
    str or None
        The first matching string from the list, preserving its original case,
        or None if there is no match.

    Examples
    --------
    >>> case_insensitive_match('foo', ['bar', 'FOO'])
    'FOO'
    """
    target_lower = target.lower()
    for string in string_list:
        if string.lower() == target_lower:
            return string
    return None



"""
Splits tabular data in the form of a list of rows into columns;
makes guesses about the role of each column for plotting purposes
(X values, Y values, and text labels).
"""


class Column(list):
    "Store a column of tabular data; record its name and whether it is numeric"

    is_quantity = True
    name = ""

    def __init__(self, *arg, **kwarg):
        pass


def is_quantity(val):
    """Is ``val`` a quantity (int, float, datetime, etc) (not str, bool)?

    Relies on presence of __sub__.
    """
    return hasattr(val, "__sub__")


class ColumnGuesserMixin(object):
    """
    plot: [x, y, y...], y
    pie: ... y
    """

    def __init__(self):
        self.keys = None

    def _build_columns(self):
        self.columns = [Column() for col in self.keys]
        for row in self:
            for col_idx, col_val in enumerate(row):
                col = self.columns[col_idx]
                col.append(col_val)
                if (col_val is not None) and (not is_quantity(col_val)):
                    col.is_quantity = False

        for idx, key_name in enumerate(self.keys):
            self.columns[idx].name = key_name

        self.x = Column()
        self.ys = []

    def _get_y(self):
        for idx in range(len(self.columns) - 1, -1, -1):
            if self.columns[idx].is_quantity:
                self.ys.insert(0, self.columns.pop(idx))
                return True

    def _get_x(self):
        for idx in range(len(self.columns)):
            if self.columns[idx].is_quantity:
                self.x = self.columns.pop(idx)
                return True

    def _get_xlabel(self, xlabel_sep=" "):
        self.xlabels = []
        if self.columns:
            for row_idx in range(len(self.columns[0])):
                self.xlabels.append(
                    xlabel_sep.join(str(c[row_idx]) for c in self.columns)
                )
        self.xlabel = ", ".join(c.name for c in self.columns)

    def _guess_columns(self):
        self._build_columns()
        self._get_y()
        if not self.ys:
            raise AttributeError("No quantitative columns found for chart")

    def guess_pie_columns(self, xlabel_sep=" "):
        """
        Assigns x, y, and x labels from the data set for a pie chart.

        Pie charts simply use the last quantity column as
        the pie slice size, and everything else as the
        pie slice labels.
        """
        self._guess_columns()
        self._get_xlabel(xlabel_sep)

    def guess_plot_columns(self):
        """
        Assigns ``x`` and ``y`` series from the data set for a plot.

        Plots use:
          the rightmost quantity column as a Y series
          optionally, the leftmost quantity column as the X series
          any other quantity columns as additional Y series
        """
        self._guess_columns()
        self._get_x()
        while self._get_y():
            pass



from sql.ggplot.aes import aes
from sql.ggplot.geom.geom import geom
from sql.ggplot.facet_wrap import facet_wrap
import matplotlib as mpl
import matplotlib.pyplot as plt
from ploomber_core.dependencies import requires


def _expand_to_multipanel_ax(figure, ax_to_clear=None):
    figure.subplots_adjust(hspace=0.7, wspace=0.5)
    if ax_to_clear:
        ax_to_clear.remove()


def _create_single_panel_ax():
    figure, ax = plt.subplots()
    axs = [ax]
    return figure, axs


@requires(["matplotlib"])
class ggplot:
    """
    Create a new ggplot
    """

    figure: mpl.figure.Figure
    axs: list

    def __init__(self, table, mapping: aes = None, conn=None, with_=None) -> None:
        self.table = table
        self.with_ = [with_] if with_ else None
        self.mapping = mapping if mapping is not None else aes()
        self.conn = conn

        figure, axs = _create_single_panel_ax()

        self.axs = axs
        self.figure = figure

    def __add__(self, other) -> "ggplot":
        """
        Add to ggplot
        """
        self._draw(other)

        return self

    def __iadd__(self, other):
        return other.__add__(self)

    def _draw(self, other) -> mpl.figure.Figure:
        """
        Draws plot
        """
        if isinstance(other, geom):
            self.geom = other
            other.draw(self)

        if isinstance(other, facet_wrap):
            _expand_to_multipanel_ax(self.figure, ax_to_clear=self.axs[0])

            values, n_rows, n_cols = other.get_facet_values(
                self.table, other.facet, with_=self.with_
            )

            for i, value in enumerate(values):
                ax_ = self.figure.add_subplot(n_rows, n_cols, i + 1)
                facet_key_val = {"key": other.facet, "value": value[0]}
                self.geom.draw(self, ax_, facet_key_val)
                handles, labels = ax_.get_legend_handles_labels()
                ax_.set_title(value[0])
                ax_.tick_params(axis="both", labelsize=7)
                # reverses legend order so alphabetically first goes on top
                ax_.legend(handles[::-1], labels[::-1], prop={"size": 10})
                if other.legend is False:
                    plt.legend("", frameon=False)
                self.axs.append(ax_)

        return self.figure

    def get_base(self, object) -> str:
        """
        Returns the base class of an object
        """
        for base in object.__class__.__bases__:
            return base.__name__



from sqlalchemy import inspect
from prettytable import PrettyTable
from sql.connection import ConnectionManager
from sql import exceptions
import math
from sql import util
from sql.store import get_all_keys
from IPython.core.display import HTML
import uuid


def _get_inspector(conn):
    if conn:
        return inspect(conn)

    if not ConnectionManager.current:
        raise exceptions.RuntimeError("No active connection")
    else:
        return inspect(ConnectionManager.current.connection_sqlalchemy)


class DatabaseInspection:
    def __repr__(self) -> str:
        return self._table_txt

    def _repr_html_(self) -> str:
        return self._table_html


class Tables(DatabaseInspection):
    """
    Displays the tables in a database
    """

    def __init__(self, schema=None, conn=None) -> None:
        inspector = _get_inspector(conn)

        self._table = PrettyTable()
        self._table.field_names = ["Name"]

        for row in inspector.get_table_names(schema=schema):
            self._table.add_row([row])

        self._table_html = self._table.get_html_string()
        self._table_txt = self._table.get_string()


def _add_missing_keys(keys, mapping):
    """
    Return a copy of `mapping` with all the missing `keys`, setting the
    value as an empty string
    """
    return {key: mapping.get(key, "") for key in keys}


# we're assuming there's one row that contains all keys, I tested this and worked fine
# my initial implementation just took all keys that appeared in "rows" but then order
# isn't preserved, which is important for user experience
def _get_row_with_most_keys(rows):
    """
    Get the row with the maximum length from the nested lists in `rows`
    """
    if not rows:
        return list()

    max_idx, max_ = None, 0

    for idx, row in enumerate(rows):
        if len(row) > max_:
            max_idx = idx
            max_ = len(row)

    if max_idx is None:
        return list()

    return list(rows[max_idx])


def _is_numeric(value):
    """Check if a column has numeric and not categorical datatype"""
    try:
        if isinstance(value, bool):
            return False
        float(value)  # Try to convert the value to float
        return True
    except (TypeError, ValueError):
        return False


def _is_numeric_as_str(column, value):
    """Check if a column contains numerical data stored as `str`"""
    try:
        if isinstance(value, str) and _is_numeric(value):
            return True
        return False
    except ValueError:
        pass


def _generate_column_styles(
    column_indices, unique_id, background_color="#FFFFCC", text_color="black"
):
    """
    Generate CSS styles to change the background-color of all columns
    with data-type mismatch.

    Parameters
    ----------
        column_indices (list): List of column indices with data-type mismatch.
        unique_id (str): Unique ID for the current table.
        background_color (str, optional): Background color for the mismatched columns.
        text_color (str, optional): Text color for the mismatched columns.

    Returns:
        str: HTML style tags containing the CSS styles for the mismatched columns.
    """

    styles = ""
    for index in column_indices:
        styles = f"""{styles}
        #profile-table-{unique_id} td:nth-child({index + 1}) {{
            background-color: {background_color};
            color: {text_color};
        }}
        """
    return f"<style>{styles}</style>"


def _generate_message(column_indices, columns):
    """Generate a message indicating all columns with a datatype mismatch"""
    message = "Columns "
    for c in column_indices:
        col = columns[c - 1]
        message = f"{message}<code>{col}</code>"
    message = (
        f"{message} have a datatype mismatch -> numeric values stored as a string."
    )
    message = f"{message} <br> Cannot calculate mean/min/max/std/percentiles"
    return message


def _assign_column_specific_stats(col_stats, is_numeric):
    """
    Assign NaN values to categorical/numerical specific statistic.

    Parameters
    ----------
        col_stats (dict): Dictionary containing column statistics.
        is_numeric (bool): Flag indicating whether the column is numeric or not.

    Returns:
        dict: Updated col_stats dictionary.
    """
    categorical_stats = ["top", "freq"]
    numerical_stats = ["mean", "min", "max", "std", "25%", "50%", "75%"]

    if is_numeric:
        for stat in categorical_stats:
            col_stats[stat] = math.nan
    else:
        for stat in numerical_stats:
            col_stats[stat] = math.nan

    return col_stats


class Columns(DatabaseInspection):
    """
    Represents the columns in a database table
    """

    def __init__(self, name, schema, conn=None) -> None:
        is_table_exists(name, schema)

        inspector = _get_inspector(conn)

        # this returns a list of dictionaries. e.g.,
        # [{"name": "column_a", "type": "INT"}
        #  {"name": "column_b", "type": "FLOAT"}]
        if not schema and "." in name:
            schema, name = name.split(".")
        columns = inspector.get_columns(name, schema) or []

        self._table = PrettyTable()
        self._table.field_names = _get_row_with_most_keys(columns)

        for row in columns:
            self._table.add_row(
                list(_add_missing_keys(self._table.field_names, row).values())
            )

        self._table_html = self._table.get_html_string()
        self._table_txt = self._table.get_string()


class TableDescription(DatabaseInspection):
    """
     Generates descriptive statistics.

     --------------------------------------
     Descriptive statistics are:

     Count - Number of all not None values

     Mean - Mean of the values

     Max - Maximum of the values in the object.

     Min - Minimum of the values in the object.

     STD - Standard deviation of the observations

     25h, 50h and 75h percentiles

     Unique - Number of not None unique values

     Top - The most frequent value

     Freq - Frequency of the top value

    ------------------------------------------
    Following statistics will be calculated for :-

    Categorical columns - [Count, Unique, Top, Freq]

    Numerical columns - [Count, Unique, Mean, Max, Min,
                         STD, 25h, 50h and 75h percentiles]

    """

    def __init__(self, table_name, schema=None) -> None:
        is_table_exists(table_name, schema)

        if schema:
            table_name = f"{schema}.{table_name}"

        conn = ConnectionManager.current

        columns_query_result = conn.raw_execute(f"SELECT * FROM {table_name} WHERE 1=0")
        if ConnectionManager.current.is_dbapi_connection:
            columns = [i[0] for i in columns_query_result.description]
        else:
            columns = columns_query_result.keys()

        table_stats = dict({})
        columns_to_include_in_report = set()
        columns_with_styles = []
        message_check = False

        for i, column in enumerate(columns):
            table_stats[column] = dict()

            # check the datatype of a column
            try:
                result = ConnectionManager.current.raw_execute(
                    f"""SELECT {column} FROM {table_name} LIMIT 1"""
                ).fetchone()

                value = result[0]
                is_numeric = isinstance(value, (int, float)) or (
                    isinstance(value, str) and _is_numeric(value)
                )
            except ValueError:
                is_numeric = True

            if _is_numeric_as_str(column, value):
                columns_with_styles.append(i + 1)
                message_check = True
            # Note: index is reserved word in sqlite
            try:
                result_col_freq_values = ConnectionManager.current.raw_execute(
                    f"""SELECT DISTINCT {column} as top,
                    COUNT({column}) as frequency FROM {table_name}
                    GROUP BY top ORDER BY frequency Desc""",
                ).fetchall()

                table_stats[column]["freq"] = result_col_freq_values[0][1]
                table_stats[column]["top"] = result_col_freq_values[0][0]

                columns_to_include_in_report.update(["freq", "top"])

            except Exception:
                pass

            try:
                # get all non None values, min, max and avg.
                result_value_values = ConnectionManager.current.raw_execute(
                    f"""
                    SELECT MIN({column}) AS min,
                    MAX({column}) AS max,
                    COUNT({column}) AS count
                    FROM {table_name}
                    WHERE {column} IS NOT NULL
                    """,
                ).fetchall()

                columns_to_include_in_report.update(["count", "min", "max"])
                table_stats[column]["count"] = result_value_values[0][2]

                table_stats[column]["min"] = round(result_value_values[0][0], 4)
                table_stats[column]["max"] = round(result_value_values[0][1], 4)

                columns_to_include_in_report.update(["count", "min", "max"])

            except Exception:
                pass

            try:
                # get unique values
                result_value_values = ConnectionManager.current.raw_execute(
                    f"""
                    SELECT
                    COUNT(DISTINCT {column}) AS unique_count
                    FROM {table_name}
                    WHERE {column} IS NOT NULL
                    """,
                ).fetchall()
                table_stats[column]["unique"] = result_value_values[0][0]
                columns_to_include_in_report.update(["unique"])
            except Exception:
                pass

            try:
                results_avg = ConnectionManager.current.raw_execute(
                    f"""
                                SELECT AVG({column}) AS avg
                                FROM {table_name}
                                WHERE {column} IS NOT NULL
                                """,
                ).fetchall()

                columns_to_include_in_report.update(["mean"])
                table_stats[column]["mean"] = format(float(results_avg[0][0]), ".4f")

            except Exception:
                table_stats[column]["mean"] = math.nan

            # These keys are numeric and work only on duckdb
            special_numeric_keys = ["std", "25%", "50%", "75%"]

            try:
                # Note: stddev_pop and PERCENTILE_DISC will work only on DuckDB
                result = ConnectionManager.current.raw_execute(
                    f"""
                    SELECT
                        stddev_pop({column}) as key_std,
                        percentile_disc(0.25) WITHIN GROUP
                        (ORDER BY {column}) as key_25,
                        percentile_disc(0.50) WITHIN GROUP
                        (ORDER BY {column}) as key_50,
                        percentile_disc(0.75) WITHIN GROUP
                        (ORDER BY {column}) as key_75
                    FROM {table_name}
                    """,
                ).fetchall()

                columns_to_include_in_report.update(special_numeric_keys)
                for i, key in enumerate(special_numeric_keys):
                    # r_key = f'key_{key.replace("%", "")}'
                    table_stats[column][key] = format(float(result[0][i]), ".4f")

            except TypeError:
                # for non numeric values
                for key in special_numeric_keys:
                    table_stats[column][key] = math.nan

            except Exception as e:
                # We tried to apply numeric function on
                # non numeric value, i.e: DateTime
                if "duckdb.BinderException" or "add explicit type casts" in str(e):
                    for key in special_numeric_keys:
                        table_stats[column][key] = math.nan

                # Failed to run sql command/func (e.g stddev_pop).
                # We ignore the cell stats for such case.
                pass

            table_stats[column] = _assign_column_specific_stats(
                table_stats[column], is_numeric
            )

        self._table = PrettyTable()
        self._table.field_names = [" "] + list(table_stats.keys())

        custom_order = [
            "count",
            "unique",
            "top",
            "freq",
            "mean",
            "std",
            "min",
            "25%",
            "50%",
            "75%",
            "max",
        ]

        for row in custom_order:
            if row.lower() in [r.lower() for r in columns_to_include_in_report]:
                values = [row]
                for column in table_stats:
                    if row in table_stats[column]:
                        value = table_stats[column][row]
                    else:
                        value = ""
                    # value = util.convert_to_scientific(value)
                    values.append(value)

                self._table.add_row(values)

        unique_id = str(uuid.uuid4()).replace("-", "")
        column_styles = _generate_column_styles(columns_with_styles, unique_id)

        if message_check:
            message_content = _generate_message(columns_with_styles, list(columns))
            warning_background = "#FFFFCC"
            warning_title = "Warning: "
        else:
            message_content = ""
            warning_background = "white"
            warning_title = ""

        current = ConnectionManager.current
        database = current.dialect
        db_driver = current._get_database_information()["driver"]

        if database and "duckdb" in database:
            db_message = ""
        else:
            db_message = f"""Following statistics are not available in
            {db_driver}: STD, 25%, 50%, 75%"""

        db_html = (
            f"<div style='position: sticky; left: 0; padding: 10px; "
            f"font-size: 12px; color: #FFA500'>"
            f"<strong></strong> {db_message}"
            "</div>"
        )

        message_html = (
            f"<div style='position: sticky; left: 0; padding: 10px; "
            f"font-size: 12px; color: black; background-color: {warning_background};'>"
            f"<strong>{warning_title}</strong> {message_content}"
            "</div>"
        )

        # Inject css to html to make first column sticky
        sticky_column_css = """<style>
 #profile-table td:first-child {
  position: sticky;
  left: 0;
  background-color: var(--jp-cell-editor-background);
  font-weight: bold;
}
 #profile-table thead tr th:first-child {
  position: sticky;
  left: 0;
  background-color: var(--jp-cell-editor-background);
  font-weight: bold; /* Adding bold text */
}
            </style>"""
        self._table_html = HTML(
            db_html
            + sticky_column_css
            + column_styles
            + self._table.get_html_string(
                attributes={"id": f"profile-table-{unique_id}"}
            )
            + message_html
        ).__html__()

        self._table_txt = self._table.get_string()


def get_table_names(schema=None):
    """Get table names for a given connection"""
    return Tables(schema)


def get_columns(name, schema=None):
    """Get column names for a given connection"""
    return Columns(name, schema)


def get_table_statistics(name, schema=None):
    """Get table statistics for a given connection.

    For all data types the results will include `count`, `mean`, `std`, `min`
    `max`, `25`, `50` and `75` percentiles. It will also include `unique`, `top`
    and `freq` statistics.
    """
    return TableDescription(name, schema=schema)


def get_schema_names(conn=None):
    """Get list of schema names for a given connection"""
    inspector = _get_inspector(conn)
    return inspector.get_schema_names()


def support_only_sql_alchemy_connection(command):
    """
    Throws a sql.exceptions.RuntimeError if connection is not SQLAlchemy
    """
    if ConnectionManager.current.is_dbapi_connection:
        raise exceptions.RuntimeError(
            f"{command} is only supported with SQLAlchemy "
            "connections, not with DBAPI connections"
        )


def _is_table_exists(table: str, conn) -> bool:
    """
    Runs a SQL query to check if table exists
    """
    if not conn:
        conn = ConnectionManager.current

    identifiers = conn.get_curr_identifiers()

    for iden in identifiers:
        if isinstance(iden, tuple):
            query = "SELECT * FROM {0}{1}{2} WHERE 1=0".format(iden[0], table, iden[1])
        else:
            query = "SELECT * FROM {0}{1}{0} WHERE 1=0".format(iden, table)
        try:
            conn.execute(query)
            return True
        except Exception:
            pass

    return False


def _get_list_of_existing_tables() -> list:
    """
    Returns a list of table names for a given connection
    """
    tables = []
    tables_rows = get_table_names()._table
    for row in tables_rows:
        table_name = row.get_string(fields=["Name"], border=False, header=False).strip()

        tables.append(table_name)
    return tables


def is_table_exists(
    table: str,
    schema: str = None,
    ignore_error: bool = False,
    conn=None,
) -> bool:
    """
    Checks if a given table exists for a given connection

    Parameters
    ----------
    table: str
        Table name

    schema: str, default None
        Schema name

    ignore_error: bool, default False
        Avoid raising a ValueError
    """
    if table is None:
        if ignore_error:
            return False
        else:
            raise exceptions.UsageError("Table cannot be None")
    if not ConnectionManager.current:
        raise exceptions.RuntimeError("No active connection")
    if not conn:
        conn = ConnectionManager.current

    table = util.strip_multiple_chars(table, "\"'")

    if schema:
        table_ = f"{schema}.{table}"
    else:
        table_ = table

    _is_exist = _is_table_exists(table_, conn)

    if not _is_exist:
        if not ignore_error:
            try_find_suggestions = not conn.is_dbapi_connection
            expected = []
            existing_schemas = []
            existing_tables = []

            if try_find_suggestions:
                existing_schemas = get_schema_names()

            if schema and schema not in existing_schemas:
                expected = existing_schemas
                invalid_input = schema
            else:
                if try_find_suggestions:
                    existing_tables = _get_list_of_existing_tables()

                expected = existing_tables
                invalid_input = table

            if schema:
                err_message = (
                    f"There is no table with name {table!r} in schema {schema!r}"
                )
            else:
                err_message = (
                    f"There is no table with name {table!r} in the default schema"
                )

            if table not in get_all_keys():
                suggestions = util.find_close_match(invalid_input, expected)
                suggestions_store = util.find_close_match(invalid_input, get_all_keys())
                suggestions.extend(suggestions_store)
                suggestions_message = util.get_suggestions_message(suggestions)
                if suggestions_message:
                    err_message = f"{err_message}{suggestions_message}"
            raise exceptions.TableNotFoundError(err_message)

    return _is_exist


def fetch_sql_with_pagination(
    table, offset, n_rows, sort_column=None, sort_order=None
) -> tuple:
    """
    Returns next n_rows and columns from table starting at the offset

    Parameters
    ----------
    table : str
        Table name

    offset : int
        Specifies the number of rows to skip before
        it starts to return rows from the query expression.

    n_rows : int
        Number of rows to return.

    sort_column : str, default None
        Sort by column

    sort_order : 'DESC' or 'ASC', default None
        Order list
    """
    is_table_exists(table)

    order_by = "" if not sort_column else f"ORDER BY {sort_column} {sort_order}"

    query = f"""
    SELECT * FROM {table} {order_by}
    OFFSET {offset} ROWS FETCH NEXT {n_rows} ROWS ONLY"""

    rows = ConnectionManager.current.execute(query).fetchall()

    columns = ConnectionManager.current.raw_execute(
        f"SELECT * FROM {table} WHERE 1=0"
    ).keys()

    return rows, columns



from sql.connection.connection import (
    ConnectionManager,
    SQLAlchemyConnection,
    DBAPIConnection,
    SparkConnectConnection,
    is_pep249_compliant,
    is_spark,
    PLOOMBER_DOCS_LINK_STR,
    default_alias_for_engine,
    ResultSetCollection,
    detect_duckdb_summarize_or_select,
)


__all__ = [
    "ConnectionManager",
    "SQLAlchemyConnection",
    "DBAPIConnection",
    "SparkConnectConnection",
    "is_pep249_compliant",
    "is_spark",
    "PLOOMBER_DOCS_LINK_STR",
    "default_alias_for_engine",
    "ResultSetCollection",
    "detect_duckdb_summarize_or_select",
]



from sql import inspect
from sql.cmd.cmd_utils import CmdParser
from sql.util import expand_args, is_rendering_required


def tables(others, user_ns):
    """
    Implementation of `%sqlcmd tables`

    This function takes in a string containing command line arguments,
    parses them to extract the schema name, and returns a list of table names
    present in the specified schema or in the default schema if none is specified.
    It also uses the kernel namespace for expanding arguments declared as variables.

    Parameters
    ----------
    others : str,
            A string containing the command line arguments.

    user_ns : dict,
        User namespace of IPython kernel

    Returns
    -------
    table_names: list
        list of tables in the schema

    """
    parser = CmdParser()

    parser.add_argument("-s", "--schema", type=str, help="Schema name", required=False)

    args = parser.parse_args(others)
    if is_rendering_required(" ".join(others)):
        expand_args(args, user_ns)

    return inspect.get_table_names(schema=args.schema)



import json
import os
import re
from pathlib import Path

import sqlparse

try:
    from ipywidgets import interact
except ModuleNotFoundError:
    interact = None
from IPython.core.magic import (
    Magics,
    cell_magic,
    line_magic,
    magics_class,
    needs_local_scope,
    no_var_expand,
)
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring
from sqlalchemy.exc import (
    OperationalError,
    ProgrammingError,
    DatabaseError,
    StatementError,
)
from traitlets.config.configurable import Configurable
from traitlets import Bool, Int, TraitError, Unicode, Dict, observe, validate
from sql.traits import Parameters

import warnings
import shlex
import sql.connection
import sql.parse
from sql.run.run import run_statements
from sql.parse import _option_strings_from_parser
from sql import display, exceptions
from sql.store import store
from sql.command import SQLCommand
from sql.magic_plot import SqlPlotMagic
from sql.magic_cmd import SqlCmdMagic
from sql._patch import patch_ipython_usage_error
from sql import util
from sql.error_handler import handle_exception
from sql._current import _set_sql_magic


from ploomber_core.dependencies import check_installed


try:
    from pandas.core.frame import DataFrame, Series
except ModuleNotFoundError:
    DataFrame = None
    Series = None


SUPPORT_INTERACTIVE_WIDGETS = ["Checkbox", "Text", "IntSlider", ""]
IF_NOT_SELECT_MESSAGE = "The query is not a SELECT type query and as \
snippets only work with SELECT queries,"
IF_SELECT_MESSAGE = "JupySQL does not support snippet expansion within CTEs yet,"


@magics_class
class RenderMagic(Magics):
    """
    %sqlrender magic which prints composed queries
    """

    @line_magic
    @magic_arguments()
    # TODO: only accept one arg
    @argument("line", default="", nargs="*", type=str)
    @argument(
        "-w",
        "--with",
        type=str,
        help="Use a saved query",
        action="append",
        dest="with_",
    )
    def sqlrender(self, line):
        args = parse_argstring(self.sqlrender, line)
        warnings.warn(
            "\n'%sqlrender' will be deprecated soon, "
            f"please use '%sqlcmd snippets {args.line[0]}' instead. "
            "\n\nFor documentation, follow this link : "
            "https://jupysql.ploomber.io/en/latest/api/magic-snippets.html#id1",
            FutureWarning,
        )
        return str(store[args.line[0]])


@magics_class
class SqlMagic(Magics, Configurable):
    """Runs SQL statement on a database, specified by SQLAlchemy connect string.

    Provides the %%sql magic."""

    autocommit = Bool(default_value=True, config=True, help="Set autocommit mode")
    autolimit = Int(
        default_value=0,
        config=True,
        allow_none=True,
        help="Automatically limit the size of the returned result sets",
    )
    autopandas = Bool(
        default_value=False,
        config=True,
        help="Return Pandas DataFrames instead of regular result sets",
    )
    autopolars = Bool(
        default_value=False,
        config=True,
        help="Return Polars DataFrames instead of regular result sets",
    )
    column_local_vars = Bool(
        default_value=False,
        config=True,
        help="Return data into local variables from column names",
    )
    displaycon = Bool(
        default_value=True, config=True, help="Show connection string after execution"
    )
    displaylimit = Int(
        default_value=10,
        config=True,
        allow_none=True,
        help=(
            "Automatically limit the number of rows "
            "displayed (full result set is still stored)"
        ),
    )
    dsn_filename = Unicode(
        default_value=str(Path("~/.jupysql/connections.ini").expanduser()),
        config=True,
        help="Path to DSN file. "
        "When the first argument is of the form [section], "
        "a sqlalchemy connection string is formed from the "
        "matching section in the DSN file.",
    )
    feedback = Int(
        default_value=1,
        config=True,
        help="Verbosity level. 0=minimal, 1=normal, 2=all",
    )
    lazy_execution = Bool(
        default_value=False,
        config=True,
        help="Whether to evaluate using ResultSet which will "
        "cause the plan to execute or just return a lazily "
        "executed plan allowing validating schemas, "
        "without expensive compute."
        "Currently only supported for Spark Connection.",
    )
    named_parameters = Parameters(
        default_value="warn",
        config=True,
        help=(
            "Allow named parameters in queries "
            "(i.e., 'SELECT * FROM foo WHERE bar = :bar')"
        ),
    )
    polars_dataframe_kwargs = Dict(
        default_value={},
        config=True,
        help=(
            "Polars DataFrame constructor keyword arguments"
            "(e.g. infer_schema_length, nan_to_null, schema_overrides, etc)"
        ),
    )
    short_errors = Bool(
        default_value=True,
        config=True,
        help="Don't display the full traceback on SQL Programming Error",
    )
    style = Unicode(
        default_value="DEFAULT",
        config=True,
        help=(
            "Set the table printing style to any of prettytable's "
            "defined styles (currently DEFAULT, MSWORD_FRIENDLY, PLAIN_COLUMNS, "
            "RANDOM, SINGLE_BORDER, DOUBLE_BORDER, MARKDOWN )"
        ),
    )

    def __init__(self, shell):
        self._store = store

        Configurable.__init__(self, config=shell.config)
        Magics.__init__(self, shell=shell)

        # Add ourself to the list of module configurable via %config
        self.shell.configurables.append(self)

    @validate("dsn_filename")
    def _valid_dsn_filename(self, proposal):
        path = Path(proposal["value"]).expanduser()
        return str(path)

    # To verify displaylimit is valid positive integer
    # If:
    #   None -> We treat it as 0 (no limit)
    #   Positive Integer -> Pass
    #   Negative Integer -> raise Error
    @validate("displaylimit")
    def _valid_displaylimit(self, proposal):
        if proposal["value"] is None:
            display.message("displaylimit: Value None will be treated as 0 (no limit)")
            return 0
        try:
            value = int(proposal["value"])
            if value < 0:
                raise TraitError(
                    "{}: displaylimit cannot be a negative integer".format(value)
                )
            return value
        except ValueError:
            raise TraitError("{}: displaylimit is not an integer".format(value))

    @observe("autopandas", "autopolars")
    def _mutex_autopandas_autopolars(self, change):
        # When enabling autopandas or autopolars, automatically disable the
        # other one in case it was enabled and print a warning
        if change["new"]:
            other = "autopolars" if change["name"] == "autopandas" else "autopandas"
            if getattr(self, other):
                setattr(self, other, False)
                display.message(
                    f"Disabled '{other}' since '{change['name']}' was enabled."
                )

    def check_random_arguments(self, line="", cell=""):
        # check only for cell magic
        if cell != "":
            tokens = shlex.split(line, posix=False)
            arguments = []

            # Iterate through the tokens to separate arguments and SQL code
            # If the token starts with "--", it is an argument
            breakLoop = False
            for token in tokens:
                if token.startswith("--") or token.startswith("-"):
                    arguments.append(token)
                    breakLoop = True
                else:
                    if breakLoop:
                        break

            declared_argument = _option_strings_from_parser(SqlMagic.execute.parser)
            for check_argument in arguments:
                if check_argument not in declared_argument:
                    raise exceptions.UsageError(
                        "Unrecognized argument(s): {}".format(check_argument)
                    )

    @no_var_expand
    @needs_local_scope
    @line_magic("sql")
    @cell_magic("sql")
    @line_magic("jupysql")
    @cell_magic("jupysql")
    @magic_arguments()
    @argument("line", default="", nargs="*", type=str, help="sql")
    @argument(
        "-l", "--connections", action="store_true", help="list active connections"
    )
    @argument("-x", "--close", type=str, help="close a session by name")
    @argument(
        "-c", "--creator", type=str, help="specify creator function for new connection"
    )
    @argument(
        "-s",
        "--section",
        type=str,
        help="section of dsn_file to be used for generating a connection string",
    )
    @argument(
        "-p",
        "--persist",
        action="store_true",
        help="create a table name in the database from the named DataFrame",
    )
    @argument(
        "-P",
        "--persist-replace",
        action="store_true",
        help="replace the DataFrame if it exists, otherwise perform --persist",
    )
    @argument(
        "-n",
        "--no-index",
        action="store_true",
        help="Do not store Data Frame index when persisting",
    )
    @argument(
        "--append",
        action="store_true",
        help=(
            "create, or append to, a table name in the database from the "
            "named DataFrame"
        ),
    )
    @argument(
        "-a",
        "--connection_arguments",
        type=str,
        help="specify dictionary of connection arguments to pass to SQL driver",
    )
    @argument("-f", "--file", type=str, help="Run SQL from file at this path")
    @argument("-S", "--save", type=str, help="Save this query for later use")
    @argument(
        "-w",
        "--with",
        type=str,
        help="Use a saved query",
        action="append",
        dest="with_",
    )
    @argument(
        "-N",
        "--no-execute",
        action="store_true",
        help="Do not execute query (use it with --save)",
    )
    @argument(
        "-A",
        "--alias",
        type=str,
        help="Assign an alias to the connection",
    )
    @argument(
        "--interact",
        type=str,
        action="append",
        help="Interactive mode",
    )
    def execute(self, line="", cell="", local_ns=None):
        """
        Runs SQL statement against a database, specified by
        SQLAlchemy connect string.

        If no database connection has been established, first word
        should be a SQLAlchemy connection string, or the user@db name
        of an established connection.

        Examples::

          %%sql postgresql://me:mypw@localhost/mydb
          SELECT * FROM mytable

          %%sql me@mydb
          DELETE FROM mytable

          %%sql
          DROP TABLE mytable

        SQLAlchemy connect string syntax examples:

          postgresql://me:mypw@localhost/mydb
          sqlite://
          mysql+pymysql://me:mypw@localhost/mydb

        """
        return self._execute(
            line=line, cell=cell, local_ns=local_ns, is_interactive_mode=False
        )

    def _execute(self, line, cell, local_ns, is_interactive_mode=False):
        """
        This function implements the cell logic; we create this private
        method so we can control how the function is called. Otherwise,
        decorating ``SqlMagic.execute`` will break when adding the ``@log_call``
        decorator with ``payload=True``

        NOTE: telemetry has been removed, we can remove this function
        """

        def interactive_execute_wrapper(**kwargs):
            for key, value in kwargs.items():
                local_ns[key] = value
            return self._execute(line, cell, local_ns, is_interactive_mode=True)

        # line is the text after the magic, cell is the cell's body

        # Examples

        # %sql {line}
        # note that line magic has no body

        # %%sql {line}
        # {cell}

        self.check_random_arguments(line, cell)

        if local_ns is None:
            local_ns = {}

        # save globals and locals so they can be referenced in bind vars
        user_ns = self.shell.user_ns.copy()
        user_ns.update(local_ns)

        command = SQLCommand(self, user_ns, line, cell)
        # args.line: contains the line after the magic with all options removed

        args = command.args

        if util.is_rendering_required(line):
            util.expand_args(args, user_ns)

        if args.section and args.alias:
            raise exceptions.UsageError(
                "Cannot use --section with --alias since the section name "
                "is automatically set as the connection alias"
            )

        is_cte = command.sql_original.strip().lower().startswith("with ")

        # only expand CTE if this is not a CTE itself
        if not is_cte:
            if args.with_:
                with_ = args.with_
            else:
                with_ = self._store.infer_dependencies(command.sql_original, args.save)
                if with_:
                    query_type = get_query_type(command.sql_original)

                    if query_type != "SELECT":
                        display.message_warning(
                            f"Your query is using the following snippets: \
{', '.join(with_)}. {IF_NOT_SELECT_MESSAGE} CTE generation is disabled"
                        )
                    else:
                        command.set_sql_with(with_)
                        display.message(
                            f"Generating CTE with stored snippets: \
{util.pretty_print(with_)}"
                        )
                else:
                    with_ = None
        else:
            query_type = get_query_type(command.sql_original)
            if args.with_:
                raise exceptions.UsageError(
                    "Cannot use --with with CTEs, remove --with and re-run the cell"
                )

            dependencies = self._store.infer_dependencies(
                command.sql_original, args.save
            )

            if dependencies:
                if query_type != "SELECT":
                    display_message = IF_NOT_SELECT_MESSAGE
                else:
                    display_message = IF_SELECT_MESSAGE
                display.message_warning(
                    f"Your query is using one or more of the following snippets: \
{', '.join(dependencies)}. {display_message}\
 CTE generation is disabled"
                )
            with_ = None

        # Create the interactive slider
        if args.interact and not is_interactive_mode:
            check_installed(["ipywidgets"], "--interactive argument")
            interactive_dict = {}
            for key in args.interact:
                interactive_dict[key] = local_ns[key]
            display.message(
                "Interactive mode, please interact with below "
                "widget(s) to control the variable"
            )
            interact(interactive_execute_wrapper, **interactive_dict)
            return

        if args.connections:
            return sql.connection.ConnectionManager.connections_table()
        elif args.close:
            return sql.connection.ConnectionManager.close_connection_with_descriptor(
                args.close
            )

        connect_arg = command.connection

        if args.section:
            connect_arg = sql.parse.connection_str_from_dsn_section(args.section, self)

        if args.connection_arguments:
            try:
                # check for string deliniators, we need to strip them for json parse
                raw_args = args.connection_arguments
                if len(raw_args) > 1:
                    targets = ['"', "'"]
                    head = raw_args[0]
                    tail = raw_args[-1]
                    if head in targets and head == tail:
                        raw_args = raw_args[1:-1]
                args.connection_arguments = json.loads(raw_args)
            except Exception as e:
                raise exceptions.ValueError(str(e)) from e
        else:
            args.connection_arguments = {}
        if args.creator:
            args.creator = user_ns[args.creator]

        # this creates a new connection or use an existing one
        # depending on the connect_arg value
        conn = sql.connection.ConnectionManager.set(
            connect_arg,
            displaycon=self.displaycon,
            connect_args=args.connection_arguments,
            creator=args.creator,
            alias=args.section if args.section else args.alias,
            config=self,
        )

        if args.persist_replace and args.append:
            raise exceptions.UsageError(
                """You cannot simultaneously persist and append data to a dataframe;
                  please choose to utilize either one or the other."""
            )
        if args.persist and args.persist_replace:
            warnings.warn("Please use either --persist or --persist-replace")
            return self._persist_dataframe(
                command.sql,
                conn,
                user_ns,
                append=False,
                index=not args.no_index,
                replace=True,
            )
        elif args.persist:
            return self._persist_dataframe(
                command.sql, conn, user_ns, append=False, index=not args.no_index
            )
        elif args.persist_replace:
            return self._persist_dataframe(
                command.sql,
                conn,
                user_ns,
                append=False,
                index=not args.no_index,
                replace=True,
            )
        if args.append:
            return self._persist_dataframe(
                command.sql, conn, user_ns, append=True, index=not args.no_index
            )

        if not command.sql:
            return

        # store the query if needed
        if args.save:
            if "-" in args.save:
                warnings.warn(
                    "Using hyphens will be deprecated soon, "
                    "please use "
                    + str(args.save.replace("-", "_"))
                    + " instead for the save argument.",
                    FutureWarning,
                )
            self._store.store(args.save, command.sql_original, with_=with_)

        if args.no_execute:
            display.message("Skipping execution...")
            return

        parameters = None
        if self.named_parameters == "disabled":
            parameters = {}
        elif self.named_parameters == "enabled":
            parameters = user_ns

        try:
            result = run_statements(conn, command.sql, self, parameters=parameters)

            if (
                result is not None
                and not isinstance(result, str)
                and self.column_local_vars
            ):
                # Instead of returning values, set variables directly in the
                # users namespace. Variable names given by column names

                if self.autopandas or self.autopolars:
                    keys = result.keys()
                else:
                    keys = result.keys
                    result = result.dict()

                if self.feedback:
                    display.message(
                        "Returning data to local variables [{}]".format(", ".join(keys))
                    )

                self.shell.user_ns.update(result)

                return None
            else:
                if command.result_var:
                    self.shell.user_ns.update({command.result_var: result})
                    if command.return_result_var:
                        return result
                    return None

                # Return results into the default ipython _ variable
                return result

        # JA: added DatabaseError for MySQL
        except (
            ProgrammingError,
            OperationalError,
            DatabaseError,
            # raised when they query has :parameters but no parameters are given
            StatementError,
        ) as e:
            # Sqlite apparently return all errors as OperationalError :/
            handle_exception(e, command.sql, self.short_errors)
        except Exception as e:
            # Handle non SQLAlchemy errors
            handle_exception(e, command.sql, self.short_errors)

    legal_sql_identifier = re.compile(r"^[A-Za-z0-9#_$]+")

    def _persist_dataframe(
        self, raw, conn, user_ns, append=False, index=True, replace=False
    ):
        """Implements PERSIST, which writes a DataFrame to the RDBMS"""
        if not DataFrame:
            raise exceptions.MissingPackageError(
                "You must install pandas to persist results: pip install pandas"
            )

        frame_name = raw.strip(";")

        # user may pass schema.dataframe (required for certain DBs
        # like Trino)
        schema_name = None
        if "." in frame_name:
            schema_frame = frame_name.split(".")
            schema_name = schema_frame[0]
            frame_name = schema_frame[1]

        # invalid identifier
        if not frame_name.isidentifier():
            raise exceptions.UsageError(
                f"Expected {frame_name!r} to be a pd.DataFrame but it's"
                " not a valid identifier"
            )

        # missing argument
        if not frame_name:
            raise exceptions.UsageError(
                "Missing argument: %sql --persist <name_of_data_frame>"
            )

        # undefined variable
        if frame_name not in user_ns:
            raise exceptions.UsageError(
                f"Expected {frame_name!r} to be a pd.DataFrame but it's undefined"
            )

        frame = user_ns[frame_name]

        if not isinstance(frame, DataFrame) and not isinstance(frame, Series):
            raise exceptions.TypeError(
                f"{frame_name!r} is not a Pandas DataFrame or Series"
            )

        # Make a suitable name for the resulting database table
        table_name = frame_name.lower()
        table_name = self.legal_sql_identifier.search(table_name).group(0)

        if replace:
            if_exists = "replace"
        elif append:
            if_exists = "append"
        else:
            if_exists = "fail"

        conn.to_table(
            table_name=table_name,
            data_frame=frame,
            if_exists=if_exists,
            index=index,
            schema=schema_name,
        )


def get_query_type(command: str):
    """
    Returns the query type of the original sql command
    """
    parsed = sqlparse.parse(command)
    query_type = parsed[0].get_type() if parsed else None
    if query_type == "UNKNOWN":
        return None
    return query_type


def set_configs(ip, file_path, alternate_path):
    """Set user defined SqlMagic configuration settings"""
    sql = ip.find_cell_magic("jupysql").__self__
    user_configs, loaded_from = util.get_user_configs(file_path, alternate_path)
    default_configs = util.get_default_configs(sql)
    table_rows = []

    success = False
    if user_configs:
        for config, value in user_configs.items():
            if config in default_configs.keys():
                default_type = type(default_configs[config])
                if isinstance(value, default_type):
                    setattr(sql, config, value)
                    table_rows.append([config, value])
                    success = True
                else:
                    display.message(
                        f"'{value}' is an invalid value for '{config}'. "
                        f"Please use {default_type.__name__} value instead."
                    )
            else:
                util.find_close_match_config(config, default_configs.keys())
        if success:
            if loaded_from is not None:
                display.message(f"Loading configurations from {loaded_from}.")
            else:
                display.message("Loading default configurations.")

    return table_rows


def load_SqlMagic_configs(ip):
    """Loads saved SqlMagic configs in pyproject.toml or ~/.jupysql/config"""

    file_path = util.find_path_from_root("pyproject.toml")
    alternate_path = Path("~/.jupysql/config").expanduser()

    table_rows = []
    try:
        table_rows = set_configs(ip, file_path, alternate_path)
    except Exception as e:
        if type(e).__name__ == "TomlDecodeError":
            display.message_warning(
                f"Could not load configuration file at {file_path}"
                f"{(' or ' + str(alternate_path)) if alternate_path else ''}"
                " (default configuration will be used).\nPlease "
                f"check that it is valid TOML: {e}"
            )
            return
        if type(e).__name__ == "ModuleNotFoundError":
            display.message(
                "The 'toml' package isn't installed. To load settings from "
                "pyproject.toml or ~/.jupysql/config, install with: "
                "pip install toml"
            )
            return
        else:
            raise

    if table_rows:
        display.message("Settings changed:")
        display.table(["Config", "value"], table_rows)


def load_ipython_extension(ip):
    """Load the magics, this function is executed when the user runs: %load_ext sql"""

    # If running within Databricks, do not use the sql magics
    if "DATABRICKS_RUNTIME_VERSION" in os.environ:
        del SqlMagic.magics["cell"]["sql"]
        del SqlMagic.magics["line"]["sql"]

    sql_magic = SqlMagic(ip)
    _set_sql_magic(sql_magic)
    ip.register_magics(sql_magic)

    load_SqlMagic_configs(ip)

    # start the default connection if the user has one in their config file
    sql.connection.ConnectionManager.load_default_connection_from_file_if_any(
        config=sql_magic
    )

    ip.register_magics(RenderMagic)
    ip.register_magics(SqlPlotMagic)
    ip.register_magics(SqlCmdMagic)

    patch_ipython_usage_error(ip)



import argparse
import sys
from sql import exceptions


class CmdParser(argparse.ArgumentParser):
    """
    Subclassing ArgumentParser as it throws a SystemExit
    error when it encounters argument validation errors.


    Now we raise a UsageError in case of argument validation
    issues.
    """

    def exit(self, status=0, message=None):
        if message:
            self._print_message(message, sys.stderr)

    def error(self, message):
        raise exceptions.UsageError(message)
